; *****************************************************************************
;
;          EMULACION DE LOS CODIGOS DE OPERACION DEL PROCESADOR 6507
;
; *****************************************************************************

; Resetea la CPU

Reset:	push	bx
        mov     bx,FEnd - RA-1          ; bx tiene la distancia entre RA y FEnd-1
ResLoop:mov     byte ptr RA[bx],0       ; inicializa su contenido en 0
	dec	bx
        jge     ResLoop                 ; ciclo para inicializar todo el rango
	pop	bx
        CargaRegs
        mov     cl,0FFh                 ; SP se inicializa en 255
        mov     bp,0FFFCh               ; obtiene la direccin de reset
        read    bl,bp                   ; desde 0FFFC,0FFFD
	inc	bp
	read	bh,bp
        mov     bp,bx                   ; y setea el PC
	ret

; *****************************************************************************
;               Configura los vectores de acuerdo al Decimal Flag
; *****************************************************************************

SetMathVectors:
	push	ax
        mov     al,[FDec]               ; AL = contenido del Decimal Flag
	cmp	al,0
        jz      SMVBinario              ; SI es 0, luego es binario

        mov     ax,offset ADCBCD        ; decimal...
        mov     [AddVector],ax          ; AX contiene el offset de ADCBCD
	mov	ax,offset SBCBCD
        mov     [SubVector],ax          ; AX contiene el offset de SBCBCD

        jmp     SMVSale
SMVBinario:
        mov     ax,offset DoSbc         ; binario...
        mov     [SubVector],ax          ; AX contiene el offset de DoSbc
	mov	ax,offset DoAdc
        mov     [AddVector],ax          ; AX contiene el offset de DoAbc
SMVSale:
	pop	ax
	ret

; *****************************************************************************
; Construye la palabra de control de estado (PSW) a partir de varios flags y
; del ltimo registro y lo deja en DL
; *****************************************************************************

GetPSW:	push	ax
        mov     dl,[RNTest]             ; dl = ltimo resultado
        and     dl,128                  ; bit 7: N (negativo)
        mov     al,[FOver]              ; bit 6: V (ovserflow)
	shl	al,6
	or	dl,al
        or      dl,020H                 ; bit 5: siempre prendido
        mov     al,[FBreak]             ; bit 4: B (break flag)
	shl	al,4
	or	dl,al
        mov     al,[FDec]               ; bit 3: flag decimal
	shl	al,3
	or	dl,al
        mov     al,[FIntDis]            ; bit 2: interrupciones deshabilitadas
	shl	al,2
	or	dl,al
        mov     dh,[RZTest]             ; bit 1: Z (cero)
	cmp	dh,0
	jnz	PSWZero
	or	dl,002H
PSWZero:or      dl,ah                   ; bit 0: C (carry)
	pop ax
	ret


; *****************************************************************************
; Cdigo del CMP
;        AL = acumulador
;        BX = apunta a los datos
;

DoCmp:  read    dl,bx           ; DL = dato ledo
        mov     bl,al           ; salva el acumulador

	xor	bh,bh
	mov	ah,0
        sahf                    ; presetea el carry
        sbb     al,dl           ; y hace el test...
	jno	CmpNO
        inc     bh              ; hubo overflow
CmpNO:  mov     [FOver],bh      ; actualiza flag de overflow
	usetest	al
	cmc
	lahf
	and	ah,1
        mov     al,bl           ; vuelve a poner el acumulador como antes
	ret
        

; *****************************************************************************
; Cdigo del SBC (binario)
;        AL = acumulador
;        AH = bit de carry
;        BX = apunta a los datos
;

DoSbc:  read    dl,bx           ; lee el dato
	xor	bh,bh
        sahf                    ; setea el bit de carry
        cmc                     ; el 6502 necesita el complemento
	sbb	al,dl
	jno	SbcNO
        inc     bh              ; cuidado el Carry!
SbcNO:  mov     [FOver],bh      ; acutliza flag de overflow
        usetest al              ; actualiza los flags RZTest y RNTest
        cmc                     ; el 6502 necesita el complemento
	lahf
	and	ah,1
	ret


; *****************************************************************************
; Cdigo del ADC (binario)
;        AL = accumulator
;        AH = bit de carry
;        BX = apunta a los datos
;

DoAdc:  read    dl,bx           ; lee el dato
        xor     bh,bh           ; presetea el bit de overflow
        sahf                    ; setea el carry
        adc     al,dl           ; realiza el ADC
	jno	AdcNO
        inc     bh              ; cuidad con el carry!
AdcNO:  mov     [FOver],bh      ; actualiza flag de carry
        usetest al              ; actualiza los flags RZTest y RNTest
	lahf
	and	ah,1
	ret



; *****************************************************************************
; Cdigo del ADC (decimal)
;        AL = acumulador
;        AH = bit de carry
;        BX = apunta a los datos

ADCBCD: read    dl,bx           ; lee el dato
	xor	bh,bh
        sahf                    ; setea el bit de carry
        adc     al,dl           ; obtiene los flags de singo y overflow
	daa
	lahf
	and	ah,1
        usetest al              ; actualiza los flags RZTest y RNTest
	ret
        


; *****************************************************************************
; Cdigo del SBC (decimal)
;       AL = acumulador
;       AH = bit de carry
;       BX = apunta a los datos

SBCBCD: read    dl,bx           ; lee el dato
	xor	bh,bh
        sahf                    ; setea el bit de carry
        cmc                     ; el 6502 necesita el complemento
        sbb     al,dl           ; obtiene los flags de singo y overflow
        das
	cmc
	lahf
	and	ah,1
	usetest	al
	ret
        


; *****************************************************************************
;
;       Cdigo de las instrucciones que soporta el 6502. No se soportan las
;       instrucciones no documentadas
;
; *****************************************************************************


; BRK -- interrupcin programada

op00:   mov     dx,bp           ; direccin de retorno en dx
        mov     bl,cl           ; bx tiene la direccin del stack (SP)
	mov	bh,1
        write   dh,bx           ; sube al stack la direccin de retorno
	dec	bl
        write   dl,bx           ;     ...2 byte
        dec     cl              ; actualiza SP
	dec	cl
	call	xop08		; push PSW
        mov     bx,0FFFEh       ; hace fetch del vector (2 bytes)
	read	dl,bx
	inc	bx
	read	dh,bx
        mov     bp,dx           ; y transfiere la direccin de ejecucin
	jmp	next


; ORA (adr,X)

op01:   indirectx
        read    dl,bx           ; lee dato
        or      al,dl           ; aplica OR en el acumulador
        jmp     ualnext         ; salta a la UAL


; ORA adr

op05:	zeropage
        zread   dl,bx           ; DL = dato ledo (del TIA o RAM)
        or      al,dl           ; aplica OR en el acumulador
        jmp     ualnext         ; salta ala UAL


; ASL adr

op06:	zeropage
        zread   dl,bx           ; DL = dato ledo (del TIA o RAM)
        DoASL                   ; aplica ASL
        rewrite dl              ; reescribe dato
	jmp	next


; PHP -- manda la pila la palabra de estado (PSW)

op08:	call	xop08
	jmp	next

xop08:  call    GetPSW          ; obtiene el PSW en DL
        mov     bl,cl           ; bx apunta a la pila
	mov	bh,1
        write   dl,bx           ; lo graba en la pila
        dec     cl              ; decrementa el puntero de la pila
	ret

; ORA dato

op09:   fetch   dl              ; obtiene el dato
        or      al,dl           ; aplica el OR al acumulador
	jmp	ualnext


; ASL A

op0a:   mov     dl,al           ; DL = acumulador
        DoASL                   ; aplica el ASL
        mov     al,dl           ; AL = nuevo valor (nuevo acumulador)
	jmp	next


; ORA adr16

op0d:   absolute                ; BX = direccin del byte que debe ser ledo
        read    dl,bx           ; DL = valor ledo
        or      al,dl           ; aplica el OR al acumulador
	jmp	ualnext


; ASL adr16

op0e:   absolute                ; BX = direccin del byte que debe ser ledo
        read    dl,bx           ; DL = valor ledo
        DoASL                   ; aplica ASL
        rewrite dl              ; reescribe el nuevo valor
op0ejn:	jmp	next


; BPL desplazamiento -- (branch if plus)

op10:   relative                ; DX = dato ledo (signo extendido)
        mov     ch,[RNTest]     ; CH = Negative Flag
        test    ch,128          ; verifica si el flag est activado
        jnz     op0ejn          ; si es negativo retorna (near return)
        jmp     DoBranch        ; realiza el salto...


; ORA (adr),Y

op11:   indirecty               ; direccionamiento indirecto
        pagechange              ; actualiza clock si cambi de pgina
        read    dl,bx           ; DL = dato ledo
        or      al,dl           ; aplica OR en el acumulador
        jmp     ualnext         ; salta a la UAL


; ORA adr,X

op15:   zeropagex               ; direccin pgina cero
        zread   dl,bx           ; DL = dato ledo (del TIA o RAM)
        or      al,dl           ; aplica OR en el acumulador
        jmp     ualnext         ; salta a la UAL


; ASL adr,X

op16:	zeropagex
        zread   dl,bx           ; DL = dato ledo (del TIA o RAM)
        DoASL                   ; realiza el ASL
        rewrite dl              ; escribe el resultado
	jmp	next


; CLC -- borra el flag de carry

op18:   xor     ah,ah           ; AH = 0 (AH tiene el flag de carry en modo ejec)
	jmp	next


; ORA adr16,Y

 op19:   absolutey              ; direccionamiento absoluto dado por Y
        pagechange              ; actualiza clock si cambi de pgina
        read    dl,bx           ; DL = dato ledo
        or      al,dl           ; aplica OR en el acumulador
	jmp	ualnext


; ORA adr16,X

op1d:   absolutex               ; direccionamiento absoluto dado por X
        pagechange              ; actualiza clock si cambi de pgina
        read    dl,bx           ; DL = dato ledo
        or      al,dl           ; aplica OR en el acumulador
	jmp	ualnext


; ASL adr16,X

op1e:   absolutex               ; direccionamiento absoluto dado por X
        read    dl,bx           ; DL = dato ledo
        DoASL                   ; aplica ASL
        rewrite dl              ; escribe en resultado
	jmp	next


; JSR adr16 -- llamada a subrutina

op20:   absolute                ; ahora en BX se encuentra la direccin de salto
	push	ax
	push	bx
	push	dx
        mov     ax,bp           ; ax es la direccin de retorno
        dec     ax              ; push retorno-1
        mov     bl,cl           ; bl es el dice al stack
	mov	bh,1
        write   ah,bx           ; push direccin de retorno
	dec	bl
	write	al,bx
        dec     cl              ; decrementa el puntero al stack en 2 bytes
	dec	cl
	pop	dx
	pop	bx
	pop	ax
        mov     bp,bx           ; transfiere el control a la direccin
        jmp     next            ; simple!


; AND (adr,X)

op21:	indirectx
        read    dl,bx           ; DL = dato ledo
        and     al,dl           ; aplica AND al acumulador
	jmp	ualnext


; BIT adr

op24:   zeropage                ; BX = pgina cero
BitTest:read    dl,bx           ; DL = dato ledo
        mov     dh,dl           ; DH = dato ledo original (lo salva)
        and     dl,al           ; este es el resultado del AND
        useztest dl             ; actualiza Z
        usentest dh             ; actualiza N
        shr     dh,6            ; el bit 6 el el flag de overflow
	and	dh,1
        mov     [FOver],dh      ; graba en el registro de la CPU emulada
	jmp	next


; AND adr

op25:	zeropage
        zread   dl,bx           ; DL = dato ledo (del TIA o RAM)
        and     al,dl           ; aplica AND al acumulador
	jmp	ualnext


; ROL adr

op26:	zeropage
        zread   dl,bx           ; DL = dato ledo (del TIA o RAM)
        DoROL                   ; hace el ROL
        rewrite dl              ; escribe el resultado
	jmp	next


; PLP -- obtiene PSW de la pila

op28:	call	xop28
	jmp	next

xop28:  inc     cl              ; incrementa el puntero a la pila
        mov     bl,cl           ; y lee el primer byte en CH
        mov     bh,1            ;
        read    ch,bx           ; CH = byte ledo
        mov     [RNTest],ch     ; actualiza el flag de negativo

        mov     ah,ch           ; bit 0: C (Carry)
	and	ah,1
        shr     ch,1            ; bit 1: Z (Cero)
	mov	dl,ch
	and	dl,1
	xor	dl,1
	mov	[RZTest],dl
        shr     ch,1            ; bit 2: INTDIS (interrupcin deshabilitada)
	mov	dl,ch
	and	dl,1
	mov	[FIntDis],dl
        shr     ch,1            ; bit 3: D (flag de decimal)
	mov	dl,ch
	and	dl,1
	mov	[FDec],dl
        shr     ch,1            ; bit 4: B (break flag)
	mov	dl,ch
	and	dl,1
	mov	[FBreak],dl
        shr     ch,2            ; bit 6: V (overflow)
	and	ch,1
	mov	[FOver],ch
        call    SetMathVectors  ; ajusta el modo decimal/binario
	ret

; AND data

op29:   fetch   dl              ; DL = dato ledo
        and     al,dl           ; aplica AND al acumulador
	jmp	ualnext


; ROL A

op2a:   mov     dl,al           ; DL = acumulador
        DoROL                   ; hace el ROL
        mov     al,dl           ; actualiza acumulador
	jmp	next


; BIT adr16

op2c:   absolute                ; BX = direccin del dato
        jmp     BitTest         ; realiza el la operacin BIT con la direccin


; AND adr16

op2d:   absolute                ; BX = direccin del dato
        read    dl,bx           ; DL = dato ledo
        and     al,dl           ; aplica AND al acumulador
	jmp	ualnext


; ROL adr16

op2e:   absolute                ; BX = direccin del dato
        read    dl,bx           ; DL = dato ledo
        DoROL                   ; realiza el ROL
        rewrite dl              ; escribe el resultado
op2ejn:	jmp	next


; BMI desplazamiento (branch if minus)

op30:   relative                ; DX = dato ledo (signo extendido)
        mov     ch,[RNTest]     ; CH = flag negativo
        test    ch,128          ; verifica el flag
        jz      op2ejn          ; si es negativo retorna (near return)
        jmp     DoBranch        ; realiza el salto


; AND (adr),Y

op31:   indirecty               ; direccionamiento indirecto
        pagechange              ; actualiza clock si cambi de pgina
        read    dl,bx           ; DL = dato ledo
        and     al,dl           ; aplica AND al acumulador
	jmp	ualnext


; AND adr,X

op35:   zeropagex               ;
        zread   dl,bx           ; DL = dato ledo (del TIA o RAM)
        and     al,dl           ; aplica AND al acumulador
	jmp	ualnext


; ROL adr,X

op36:	zeropagex
        zread   dl,bx           ; DL = dato ledo (del TIA o RAM)
        DoROL                   ; Aplica ROL
        rewrite dl              ; escribe el resultado
	jmp	next


; SEC -- activa el carry

op38:   mov     ah,1            ; AH = carry en el procesador emulado
	jmp	next


; AND adr16,Y

op39:	absolutey
        pagechange              ; actualiza clock si cambi de pgina
        read    dl,bx           ; DL = dato ledo
        and     al,dl           ; aplica AND al acumulador
	jmp	ualnext


; AND adr16,X

op3d:   absolutex               ; BX tiene la direccin del dato
        pagechange              ; actualiza clock si cambi de pgina
        read    dl,bx           ; DL = dato ledo
        and     al,dl           ; aplica AND al acumulador
	jmp	ualnext


; ROL adr16,X

op3e:   absolutex               ; BX tiene la direccin del dato
        read    dl,bx           ; DL = dato ledo
        DoROL                   ; aplica el ROL
        rewrite dl              ; escribe el resultado
	jmp	next


; RTI -- retorna de una interrupcin

op40:   call    xop28           ; obtiene de la pila la PSW
        inc     cl              ; icual que el retono pero no incrementa el PC
        mov     bl,cl           ; BX apuntar a la direccin de retorno en la pila
        inc     cl              ; SP se incremento en dos
	mov	bh,1
        read    dl,bx           ; lee la direccin de retorno en BX
	inc	bl
	read	dh,bx
        mov     bp,dx           ; y retorna...
	jmp	next


; EOR (adr,X)

op41:   indirectx
        read    dl,bx           ; DL = dato ledo
        xor     al,dl           ; aplica OR al acumulador
	jmp	ualnext


; EOR adr

op45:	zeropage
        zread   dl,bx           ; DL = dato ledo (del TIA o RAM)
        xor     al,dl           ; aplica OR al acumulador
	jmp	ualnext


; LSR adr

op46:	zeropage
        zread   dl,bx           ; DL = dato ledo (del TIA o RAM)
        DoLSR                   ; realiza LSR
        rewrite dl              ; escribe el resultado
	jmp	next


; PHA -- enva a la pila el acumulador (push accumulator)

op48:   mov     bl,cl           ; BX = direcci en la pila
	mov	bh,1
        write   al,bx           ; escribe el acumulador en la pila
        dec     cl              ; decrementa la pila
	jmp	next


; EOR data

op49:   fetch   dl              ; DL = dato ledo
        xor     al,dl           ; aplica XOR al acumulador
	jmp	ualnext


; LSR A

op4a:   mov     dl,al           ; DL = acumulador
        DoLSR                   ; realiza LSR
        mov     al,dl           ; AL = resultado de la operacion (actualiza acumulador)
	jmp	next


; JMP adr16

op4c:   absolute                ; obtiene la direccin
        mov     bp,bx           ; salta a ella (actualizando el PC)
	jmp	next


; EOR adr16

op4d:   absolute                ; BX = direccin del dato a leer
        read    dl,bx           ; DL = dato ledo
        xor     al,dl           ; aplica XOR al acumulador
	jmp	ualnext


; LSR adr16

op4e:   absolute                ; BX = direccin del dato a leer
        read    dl,bx           ; DL = dato ledo
        DoLSR                   ; realiza LSR
        rewrite dl              ; escribe el resultado
op4ejn:	jmp	next


; BVC disp -- branch if	overflow clear

op50:	relative
        mov     ch,[FOver]      ; CH = flag de overflow
        cmp     ch,0            ; verifica si es 0
        jnz     op4ejn          ; si es cero retorna (near return)
        jmp     DoBranch        ; realiza el salto


; EOR (adr),Y

op51:	indirecty
        pagechange              ; actualiza clock si cambi de pgina
        read    dl,bx           ; DL = dato ledo
        xor     al,dl           ; aplica XOR al acumulador
	jmp	ualnext


; EOR adr,X

op55:	zeropagex
        zread   dl,bx           ; DL = dato ledo (del TIA o RAM)
        xor     al,dl           ; aplica XOR al acumulador
	jmp	ualnext


; LSR adr,X

op56:	zeropagex
        zread   dl,bx           ; DL = dato ledo (del TIA o RAM)
        DoLSR                   ; realiza LSR
        rewrite dl              ; escribe el resultado
	jmp	next


; CLI -- deshabilita interrupciones

op58:   mov     [FIntDis],0     ; actualiza flag de interrupcin
	jmp	next


; EOR adr16,Y

op59:	absolutey
        pagechange              ; actualiza clock si cambi de pgina
        read    dl,bx           ; DL = dato ledo
        xor     al,dl           ; aplica XOR al acumulador
	jmp	ualnext


; EOR adr16,X

op5d:   absolutex
        pagechange              ; actualiza clock si cambi de pgina
        read    dl,bx           ; DL = dato ledo
        xor     al,dl           ; aplica XOR al acumulador
	jmp	ualnext


; LSR adr16,X

op5e:	absolutex
        zread   dl,bx           ; DL = dato ledo (del TIA o RAM)
        DoLSR                   ; realiza LSR
        rewrite dl              ; escribe resultado
	jmp	next


; RTS -- vuelve de una subrutina

op60:	inc	cl
        mov     bl,cl           ; bx apuntar a la direccin de retorno en stack
        inc     cl              ; sp sube 2 bytes (todo bien!)
	mov	bh,1
        read    dl,bx           ; lee direccin de retorno en dx (dl primer byte)
	inc	bl
        read    dh,bx           ; lee segundo byte de la direccin de retorno
        mov     bp,dx           ; y finalmente retorna...
        inc     bp              ; bp = prxima instruccin
	jmp	next


; ADC (adr,X)

op61:	indirectx
        call    [AddVector]     ; llama a rutina de suma segn flag decimal/binario
	jmp	next


; ADC adr

op65:	zeropage
        call    [AddVector]     ; llama a rutina de suma segn flag decimal/binario
	jmp	next


; ROR adr

op66:	zeropage
        zread   dl,bx           ; DL = dato ledo (del TIA o RAM)
        DoROR                   ; realiza ROR
        rewrite dl              ; escribe el resultado
	jmp	next


; PLA -- obtiene acumulador de la pila

op68:   inc     cl              ; incrementa SP
        mov     bl,cl           ; construye la direccin de lectura
	mov	bh,1
        read    al,bx           ; lee la direccin
        jmp     ualnext         ; configura los flags...


; ADC data

op69:   mov     bx,bp           ; BX = PC
        inc     bp              ; BP tiene el prxima dato una vez aumentado
        call    [AddVector]     ; llama a rutina de suma segn flag decimal/binario
	jmp	next


; ROR A

op6a:   mov     dl,al           ; DL = acumulador
        DoROR                   ; realiza el ROR
        mov     al,dl           ; actualiza el acumulador con el resultado
	jmp	next


; JMP (adr16)

op6c:   absolute                ; BX = direccin del dato que contiene el salto
        read    dl,bx           ; obtiene la direccin de salto
        inc     bx              ;
        read    dh,bx           ; DX = direccin leda
        mov     bp,dx           ; actualiza el PC (realiza el salto)
        jmp     next


; ADC adr16

op6d:   absolute                ; BX = direccin del dato
        call    [AddVector]     ; llama a rutina de suma segn flag decimal/binario
	jmp	next


; ROR adr16

op6e:   absolute                ; BX = direccin del salto
        read    dl,bx           ; DL = dato ledo
        DoROR                   ; realiza el ROR
        rewrite dl              ; escribe el dato
	jmp	next


; BVS desplazamiento -- salta si el bit de overflow est encendido

op70:	relative
        mov     ch,[FOver]      ; CH = overflow
        cmp     ch,0            ; es 0 ?
        jz      op71jn          ; si no est activado retorna (near return)
        jmp     DoBranch        ; realiza el salto


; ADC (adr),Y

op71:	indirecty
        pagechange              ; actualiza clock si cambi de pgina
        call    [AddVector]     ; llama a rutina de suma segn flag decimal/binario
op71jn:	jmp	next


; ADC adr,X

op75:	zeropagex
        call    [AddVector]     ; llama a rutina de suma segn flag decimal/binario
	jmp	next


; ROR adr,X

op76:	zeropagex
        zread   dl,bx           ; DL = dato ledo (del TIA o RAM)
        DoROR                   ; realiza el ROR
        rewrite dl              ; escribe el dato
	jmp	next


; SEI -- deshabilita las interrupciones

op78:   mov     [FIntDis],1     ; flag de interrupciones deshabilitada
	jmp	next


; ADC adr16,Y

op79:	absolutey
        pagechange              ; actualiza clock si cambi de pgina
        call    [AddVector]     ; llama a rutina de suma segn flag decimal/binario
	jmp	next


; ADC adr16,X

op7d:	absolutex
        pagechange              ; actualiza clock si cambi de pgina
        call    [AddVector]     ; llama a rutina de suma segn flag decimal/binario
	jmp	next


; ROR adr16,X

op7e:	absolutex
        read    dl,bx           ; DL = dato ledo
        DoROR                   ; realiza el ROR
        rewrite dl              ; escribe resultado
	jmp	next


; STA adr,X

op81:	indirectx
        write   al,bx           ; graba el acumulador en BX
	jmp	next


; STY adr

op84:	zeropage
        mov     dl,[RY]         ; DL = Y
        zwrite  dl,bx           ; graba DL en BX (BX es TIA o RAM)
	jmp	next


; STA adr

op85:	zeropage
        zwrite  al,bx           ; graba acumulador en BX (BX es TIA o RAM)
	jmp	next


; STX adr

op86:	zeropage
        mov     dl,[RX]         ; DL = X
        zwrite  dl,bx           ; graba DL en BX (BX es TIA o RAM)
	jmp	next


; DEY -- decrementa Y

op88:   mov     bl,[RY]
        dec     bl
        mov     [RY],bl         ; Y = Y - 1
        usetest bl              ; actualiza flags N y Z
	jmp	next


; TXA -- transfiere X al acumulador

op8a:   mov     al,[RX]         ; A = X
	jmp	ualnext


; STY adr16

op8c:   absolute                ; BX = direccin donde se debe escribir el dato
        mov     dl,[RY]         ; DL = Y
        write   dl,bx           ; graba registro en BX
	jmp	next


; STA adr16

op8d:   absolute                ; BX = direccin donde se debe escribir el dato
        write   al,bx           ; graba acumulador en BX
	jmp	next


; STX adr16

op8e:   absolute                ; BX = direccin donde se debe escribir el dato
        mov     dl,[RX]         ; DL = X
        write   dl,bx           ; graba registro en BX
	jmp	next


; BCC desplzamiento -- salta si el carry est desactivado

op90:	relative
	cmp	ah,0
        jnz     op91jn          ; si no est activado no realiza el salto
        jmp     DoBranch        ; realiza el salto


; STA (adr),Y

op91:   indirecty               ; BX = direccin donde se debe escribir el dato
        write   al,bx           ; escribe acumulador en BX
op91jn:	jmp	next


; STY adr,X

op94:	zeropagex
        mov     dl,[RY]         ; DL = Y
        zwrite  dl,bx           ; escibe registro en BX (BX es TIA o RAM)
	jmp	next


; STA adr,X

op95:	zeropagex
        zwrite  al,bx           ; escribe acumulador en BX (BX es TIA o RAM)
	jmp	next


; STX adr,Y

op96:	zeropagey
        mov     dl,[RX]         ; DL = X
        zwrite  dl,bx           ; escibe registro en BX (BX es TIA o RAM)
	jmp	next


; TYA -- transifere Y al acumulador

op98:   mov     al,[RY]         ; A = Y
	jmp	ualnext


; STA adr16,Y

op99:   absolutey               ; BX = direccin donde se debe escribir el dato
        write   al,bx           ; escribe acumulador en BX
	jmp	next


; TXS -- transfiere X a S

op9a:   mov     cl,[RX]         ; CL = X (CL representa al SP cuando ejecuta la CPU)
	jmp	next


; STA adr16,X

op9d:   absolutex               ; BX = direccin donde se debe escribir el dato
        write   al,bx           ; escribe acumulador en BX
	jmp	next


; LDY data

opa0:   fetch   bl              ; BL = dato
        mov     [RY],bl         ; Y = dato ledo
        usetest bl              ; actualiza flags
	jmp	next


; LDA (adr,X)

opa1:	indirectx
        read    al,bx           ; AL = dato ledo (AL es el acumulador)
        jmp     ualnext


; LDX data

opa2:   fetch   bl              ; BL = dato
        mov     [RX],bl         ; X = dato ledo
        usetest bl              ; actualiza flags
	jmp	next


; LDY adr

opa4:	zeropage
        zread   dl,bx           ; DL = dato ledo (del TIA o RAM)
        usetest dl              ; actualiza flags
        mov     [RY],dl         ; Y = dato ledo
	jmp	next


; LDA adr

opa5:	zeropage
        zread   al,bx           ; AL = dato ledo (del TIA o RAM)
	jmp	ualnext


; LDX adr

opa6:	zeropage
        zread   dl,bx           ; DL = dato ledo (del TIA o RAM)
        usetest dl              ; actualiza flags
        mov     [RX],dl         ; X = dato ledo
	jmp	next


; TAY -- transfiere el acumulador al registro Y

opa8:   mov     [RY],al         ; Y = A
	jmp	ualnext


; LDA data

opa9:   fetch   al              ; actualiza el acumulador con el dato ledo
	jmp	ualnext


; TAX -- transfiere el acumulador al registro X

opaa:   mov     [RX],al         ; X = acumulador
	jmp	ualnext


; LDY adr16

opac:   absolute                ; BX = direccin del dato a leer
        read    dl,bx           ; DL = dato ledo
        usetest dl              ; actualiza flags
        mov     [RY],dl         ; Y = dato ledo
	jmp	next


; LDA adr16

opad:   absolute                ; BX = direccin del dato a leer
        read    al,bx           ; AL = dato ledo
	jmp	ualnext


; LDX adr16

opae:   absolute                ; BX = direccin del dato a leer
        read    dl,bx           ; DL = dato ledo
        usetest dl              ; actualiza flags
        mov     [RX],dl         ; X = dato ledo
opaejn:	jmp	next


; BCS desplazamiento -- salta si el carry est activado

opb0:	relative
        cmp     ah,0            ; verifica si el carry est activado
        jz      opaejn          ; no, retorna (near return)
        jmp     DoBranch        ; ejecuta el salto


; LDA (adr),Y

opb1:	indirecty
        pagechange              ; actualiza clock si cambi de pgina
        read    al,bx           ; AL = registro ledo
	jmp	ualnext


; LDY adr,X

opb4:	zeropagex
        zread   dl,bx           ; DL = dato ledo (del TIA o RAM)
        usetest dl              ; actualiza flags
        mov     [RY],dl         ; Y = dato ledo
	jmp	next


; LDA adr,X

opb5:	zeropagex
        zread   al,bx           ; AL = dato ledo (del TIA o RAM)
	jmp	ualnext


; LDX adr,Y

opb6:	zeropagey
        zread   dl,bx           ; DL = dato ledo (del TIA o RAM)
        usetest dl              ; actualiza flags
        mov     [RX],dl         ; X = dato ledo
	jmp	next


; CLV -- borra el overflow

opb8:   mov     [FOver],0       ; actualiza flag de overflow
	jmp	next


; LDA adr16,Y

opb9:	absolutey
        pagechange              ; BX = direccin del dato que debe ser ledo
        read    al,bx           ; AL = dato ledo (actualiza acumulador)
	jmp	ualnext


; TSX -- transfiere el SP al registro X

opba:   mov     [RX],cl         ; X = SP
        usetest cl              ; actualiza flags
	jmp	next


; LDY adr16,X

opbc:	absolutex
        pagechange              ; BX = direccin del dato que debe ser ledo
        read    dl,bx           ; DL = dato ledo
        usetest dl              ; actualiza flags
        mov     [RY],dl         ; Y = dato ledo
	jmp	next


; LDA adr16,X

opbd:	absolutex
        pagechange              ; actualiza clock si cambi de pgina
        read    al,bx           ; AL = dato ledo (actualiza acumulador)
	jmp	ualnext


; LDX adr16,Y

opbe:	absolutey
        pagechange              ; BX = direccin del dato que debe ser ledo
        read    dl,bx           ; DL = dato ledo
        usetest dl              ; actualiza flags
        mov     [RX],dl         ; X = dato ledo
	jmp	next


; CPY data

opc0:   mov     bx,bp           ; BP apunta al dato ahora
        inc     bp              ; ...prximo byte
        mov     dl,[RY]         ; DL = Y (valor a comparar)
        jmp     CompDL          ; realiza la comparacin


; CMP (adr,X)

opc1:   indirectx               ; DL contiene el valor a compara
        call    DoCmp           ; realiza la comparacin
	jmp	next


; CPY adr

opc4:	zeropage
        mov     dl,[RY]         ; DL = Y (valor a comparar)
        jmp     CompDL          ; realiza la comparacin


; CMP adr

opc5:   zeropage                ; BX = direccin del dato a comparar
        call    DoCmp           ; realiza la comparacin
	jmp	next


; DEC adr

opc6:   zeropage                ; BX = direccin del dato que debe ser ledo
        zread   dl,bx           ; DL = dato ledo (del TIA o RAM)
        dec     dl              ; DL = DL - 1
        rewrite dl              ; graba el resultado
        usetest dl              ; actualiza los flags
	jmp	next


; INY -- incrementa registro Y

opc8:   mov     bl,[RY]         ; BL = Y
        inc     bl              ; BL = BL + 1
        mov     [RY],bl         ; actualiza registro Y
        usetest bl              ; actualiza flags
	jmp	next


; CMP data

opc9:   mov     bx,bp           ; BX = PC
        inc     bp              ; PC apunta al prximo byte
        call    DoCmp           ; realiza la compracin
	jmp	next


; DEX -- decrementa registro X

opca:   mov     bl,[RX]         ; BL = X
        dec     bl              ; BL = BL - 1
        mov     [RX],bl         ; actualiza registro X
        usetest bl              ; actualiza flags
	jmp	next


; CPY adr16

opcc:   absolute
        mov     dl,[RY]         ; DL = Y (dato a ser comparado)
        jmp     CompDL          ; realiza la comparacin


; CMP adr16

opcd:   absolute                ; BX = direccin del dato a comparar
        call    DoCmp           ; realiza la comparacin
	jmp	next


; DEC adr16

opce:	absolute
        read    dl,bx           ; DL = dato ledo
        dec     dl              ; DL = DL - 1
        rewrite dl              ; graba resultado
        usetest dl              ; acutaliza flags
	jmp	next


; BNE disp -- branch if	non-zero

opd0:	relative
        mov     bl,[RZTest]     ; BL = Z
        cmp     bl,0            ; flag activado ?
        jz      opd1jn          ; no.. retorna (near return)
        jmp     DoBranch        ; realiza el salto


; CMP (adr),Y

opd1:	indirecty
        pagechange              ; BX = direccin del dato a comparar
        call    DoCmp           ; realiza la comparacin
opd1jn:	jmp	next


; CMP adr,X

opd5:   zeropagex               ; BX = direccin del dato a comparar
        call    DoCmp           ; realiza la comparacin
        jmp     next


; DEC adr,X

opd6:   zeropagex               ; BX = direccin del dato a leer
        zread   dl,bx           ; DL = dato ledo (del TIA o RAM)
        dec     dl              ; DL = DL - 1
        rewrite dl              ; reescribe el resultado en la direccin BX
        usetest dl              ; actualiza flags
	jmp	next


; CLD -- desactiva modo decimal

opd8:   mov     [FDec],0        ; actualiza flag decimal
        call    SetMathVectors  ; ajusta modo decimal/binario
	jmp	next


; CMP adr16,Y

opd9:	absolutey
        pagechange              ; BX = direccin del dato a comparar
        call    DoCmp           ; realiza la comparacin
	jmp	next


; CMP adr16,X

opdd:	absolutex
        pagechange              ; BX = direccin del dato a comparar
        call    DoCmp           ; realiza la comparacin
	jmp	next


; DEC adr16,X

opde:   absolutex               ; direccin del dato que debe ser ledio
        read    dl,bx           ; DL = dato ledo
        dec     dl              ; DL = DL - 1
        rewrite dl              ; graba el resultado
        usetest dl              ; actualiza flags
	jmp	next


; CPX data

ope0:   mov     bx,bp           ; BX = PC
        inc     bp              ; avanza el PC al dato
        mov     dl,[RX]         ; DL = X
        jmp     CompDL          ; realiza la comparacin


; SBC adr,X

ope1:	indirectx
        call    [SubVector]     ; llama a rutina de resta segn flag decimal/binario
	jmp	next


; CPX adr

ope4:	zeropage
        mov     dl,[RX]         ; DL = X
        jmp     CompDL          ; realiza la comparacin


; SBC adr

ope5:	zeropage
        call    [SubVector]     ; llama a rutina de resta segn flag decimal/binario
	jmp	next


; INC adr

ope6:   zeropage                ; BX = direccin del dato a leer
        zread   dl,bx           ; DL = dato ledo (del TIA o RAM)
        inc     dl              ; DL = DL + 1
        rewrite dl              ; graba el resultado
        usetest dl              ; actualiza flags
	jmp	next


; INX -- incrementa el registro X

ope8:   mov     bl,[RX]         ; BL = X
        inc     bl              ; BL = BL + 1
        mov     [RX],bl         ; actualiza el valor del registro X
        usetest bl              ; actualiza flags
	jmp	next


; SBC data

ope9:   mov     bx,bp           ; BX = BP
        inc     bp              ; BP = BP + 1
	call	[SubVector]
	jmp	next


; NOP

opea:	jmp	next


; CPX adr16

opec:	absolute
	mov	dl,[RX]		; dl is	X Register
	jmp	CompDL


; SBC adr16

oped:	absolute
        call    [SubVector]     ; llama a rutina de resta segn flag decimal/binario
	jmp	next


; INC adr16

opee:   absolute                ; BX = direccin del dato que debe ser ledo
        read    dl,bx           ; DL = dato ledo
        inc     dl              ; DL = DL + 1
        rewrite dl              ; graba el resultado
        usetest dl              ; actualiza flags
	jmp	next


; BEQ desplazamiento -- salta si es cero

opf0:	relative
        mov     bl,[RZTest]     ; BL = Z
        cmp     bl,0            ; flag Z activado
        jnz     opf1jn          ; no, retorna (near return)
        jmp     DoBranch        ; realiza el salto


; SBC (adr),Y

opf1:	indirecty
        pagechange              ; actualiza clock si cambi de pgina
        call    [SubVector]     ; llama a rutina de resta segn flag decimal/binario
opf1jn:	jmp	next


; SBC adr,X

opf5:	zeropagex
        call    [SubVector]     ; llama a rutina de resta segn flag decimal/binario
	jmp	next


; INC adr,X

opf6:   zeropagex               ; BX = direccin del dato que debe ser ledo
        zread   dl,bx           ; DL = dato ledo (del TIA o RAM)
        inc     dl              ; DL = DL + 1
        rewrite dl              ; graba el resultado
        usetest dl              ; actualiza flags
	jmp	next


; SED -- activa modo decimal

opf8:   mov     [FDec],1        ; actualiza flag decimal
        call    SetMathVectors  ; adjusta modo decimal/binario
	jmp	next


; SBC adr16,Y

opf9:	absolutey
        pagechange              ; actualiza clock si cambi de pgina
        call    [SubVector]     ; llama a rutina de resta segn flag decimal/binario
	jmp	next


; SBC adr16,X

opfd:	absolutex
        pagechange              ; actualiza clock si cambi de pgina
        call    [SubVector]     ; llama a rutina de resta segn flag decimal/binario
	jmp	next


; INC adr16,X

opfe:   absolutex               ; BX = direccin del dato que debe ser ledo
        zread   dl,bx           ; DL = dato ledo (del TIA o RAM)
        inc     dl              ; DL = DL + 1
        rewrite dl              ; graba el resultado
        usetest dl              ; actualiza flags
	jmp	next


; *****************************************************************************
;
;                         Maneja saltos relativos
;
;       asume BP = PC y que la macro "relative" puso la direccin en DX
;
; *****************************************************************************

DoBranch:
        mov     bx,bp                   ; BH es la pgina actual
        add     dx,bp                   ; trabaja con la direccin de retorno
        mov     bp,dx                   ; y fija el PC

        inc     [RClock]                ; se requiere una ciclo extra
	inc	[RCycles]
        cmp     bh,dh                   ; cambi la pagina ?
	je	dbjn
        inc     [RCycles]               ; salto efectivo... otro ciclo extra
	inc	[RClock]
dbjn:	jmp	next

RelExit:ret


; *****************************************************************************
;
;                           Compara DL con [BP]
;
; *****************************************************************************

CompDL: mov     dh,al                   ; DH = A (salva el acumulador)
        push    dx
        mov     al,dl                   ; DL para ser comparado
        call    DoCmp                   ; realiza la comparacin
        pop     dx                      ; restaura el acumulador
        mov     al,dh
	jmp	next


;*****************************************************************************
; controladores de registros y memoria de la CPU -- utilizado por el emulador
;
;
; Procesa el timer. Ver especificaciones en el manual!
;
;
;
;

Timer1: movzx   dx,[RCycles]            ; timer normal, decrementa en 1 por
        sub     [Timer],dx              ; ciclo de procesador
Timer0:
	ret
                                        ; este decrece en diferentes formas
Timer2: movzx   dx,[RCycles]            ; resta los ciclos de tickcount
	sub	[TickCnt],dx
        jl      DecTimer                ; si es < 0, decrementa el timer
	ret
DecTimer:
        mov     dx,[TRate]              ; resetea el contador de ticks
	add	[TickCnt],dx
        dec     [Timer]                 ; decrementa el timer
	mov	dx,[Timer]
        cmp     dx,0FFFFh               ;
        je      SetMode1                ; si es zero cuenta regresiva en mode 1
	ret


;*
;* Esto es generado escribiendo a las direcciones $0294 .. $0297, registros de tiempo del RIOT
;* SI tiene la direccin al comienzo. [WByte] es el valor del timer
;*

SetTimer:
        push    ax                      ; setea el timer
        movzx   ax,[WByte]              ; obtiene el valor a escribir
        mov     [Timer],ax              ; y lo escribe!
        pop     ax
        cmp     si,0294h                ; es el timer de 1 clock por vez ?
        je      SetMode1                ; de ser as, va al Modo 1

        push    ax                      ; es el timer de Modo n
        push    bx
        mov     bx,si                   ; trabaja el offset
        sub     bx,0295h
        add     bx,bx
        mov     ax,TimerCount[bx]       ; obtiene el valor a contat
        mov     [TRate],ax              ; setea el tick rate
        movzx   bx,[RCycles]
        add     ax,bx                   ; suma el nmero de ciclos de esta instr.
                                        ; porque despus de escribir ser decrementado
        mov     [TickCnt],ax            ; setea el contador de Ticks
        mov     [TimerVec],offset Timer2; nueva direccin de llamada a Timer2 en TimerVec
        pop     bx                      ;
        pop     ax
        ret

SetMode1:                               ; nueva direccin de llamada a Time12 en TimerVec
	mov	[TimerVec],offset Timer1
	ret


;*
;* LECTURA banco de memoria (bank switcher) (SI tiene la posicin direccionado)
;* Para el cambio de banco (bank) existen una direcciones especiales dentro del
;*   espacio de direccionamiento, el cual con slo direccionar, hace referencia
;*   al banco sobre el cual se desea trabajar. Este rea de cambio vara segn
;*   el tamao del cartucho con el cual se est trabajando.
;*

RBank4: and	si,01fffh
        cmp     si,1000h                ; si no es ROM, lee de hardware
	jb	ReadHardware
	add	si,offset CartRom - 1000h
	ret
        
RBank8: and	si,01fffh
        cmp     si,1000h                ; si no es ROM, lee de hardware
	jb	ReadHardware
        cmp     si,1FF8h                ; si no est en el rea de cambio (switch)...
        jb      RBank8NoChange          ; no hay cambio
	cmp	si,1FFAh
	ja	RBank8NoChange
        mov     [RomBank],0             ; bank 0 seleccionado
	cmp	si,1FF8h
	je	RBank8NoChange
        mov     [RomBank],1000h         ; bank 1 seleccionado
RBank8NoChange:
	add	si,[RomBank]
	add	si,offset CartRom - 1000h
	ret
        
ReadHardware:
        cmp     si,0200h                ; es el RIOT ???
	jge	ReadRIOT
ReadHardwarePage0:
        and     si,0FFh                 ; debe ser el TIA o la RAM
        cmp     si,080h                 ; es el TIA ???
	jl	ReadTIA
        add     si,offset RiotRam-128   ; apuntamos si al rea del RIOTRAM
	ret
					        
RBank16: and	si,01fffh
        cmp     si,1000h                ; si no es ROM, lee de hardware
	jb	ReadHardware
        cmp     si,1FF6h                ; si no est en el rea de cambio (switch)...
        jb      RBank16NoChange         ; no hay cambio
	cmp	si,1FFAh
	ja	RBank16NoChange

        push    ax                      ; switch en ROM de 16k
	mov	ax,si
        sub     ax,1FF6h                ; es el banco 0,1,2,3
        shl     ax,12                   ; es 0,4k,8k,12k
	mov	[RomBank],ax
	pop	ax

RBank16NoChange:
	add	si,[RomBank]
	add	si,offset CartRom - 1000h
	ret
        

; *****************************************************************************
;     Mapeo de Memoria - Lectura - Para reas no pertenientes a la ROM, SI
;     tiene la direccin. A la salida ds:[si] apunta al data requerido.
; *****************************************************************************

ReadRIOT:
        and     si,02FFh                ; RIOT est en las direcciones 2xx 3xx
        cmp     si,0284h                ; es el timer del RIOT ???
	jne	NotRTimer
        mov     si,offset Timer         ; SI apuntar al timer
	ret
NotRTimer:                              ; todos los otros registros del RIOT...
        cmp     si,0280h                ; es el Port A (Controladores de Mano) ?
	je	ReadPortA
        cmp     si,0282h                ; es el Port B (Switches de Consola) ?
	je	ReadPortB
	mov	si,offset RetFF
	ret
ReadPortB:                              ; lee los switches de la consola (port b)
	mov	si,offset IOPortB
	ret
ReadPortA:                              ; lee los conotroladores de Mano (port a)
	mov	si,offset IOPortA
	ret

ReadTIA:and	si,0Fh
        cmp     si,08h                  ; leyendo registros de colisin ???
	jl	ReadCollision
	cmp	si,0Eh
	jl	ReadInputLatches
        mov     si,offset Ret0F         ; lee el TIA (default)
	ret
        

ReadCollision:                          ; lee registros de colisin (colission latch)
	push	ax
	push	cx

	mov	ax,[TIACollide]
	mov	cx,si			; cx = address
        add     cl,cl                   ; shift a derecha 2 x address
        shr     ax,cl                   ; y lo hace....
        and     ax,3                    ; ax ahora es los bits 7,6 de colisin
        shl     al,6                    ; los ponemos de vuelta en los bits 7 y 6
        mov     [RetWd],al              ; graba word para resultado (retorno)

	pop	cx
	pop	ax
	mov	si,offset RetWd
	ret

ReadInputLatches:                       ; lee registros de entrada (input latch)
	add	si,offset DumpPorts-8
	ret
        


;*
;* ESCRITURA banco de memoria (bank switcher) (SI tiene la posicin direccionado)
;* Para el cambio de banco (bank) existen una direcciones especiales dentro del
;*   espacio de direccionamiento, el cual con slo direccionar, hace referencia
;*   al banco sobre el cual se desea trabajar. Este rea de cambio vara segn
;*   el tamao del cartucho con el cual se est trabajando.
;*

; Banco de 4k

WBank4: and     si,01fffh
	cmp	si,1000h
	jb	WriteHardware
	ret
        
; Banco de 8k

WBank8: and     si,01fffh
        cmp     si,1000h                ; si no es ROM, escribe de hardware
	jb	WriteHardware
        cmp     si,1FF8h                ; si no est en el rea de cambio (switch)...
        jb      Bank8SinCambio          ; no hay cambio
	cmp	si,1FFAh
        ja      Bank8SinCambio
        mov     [RomBank],0             ; bank 0 seleccionado
	cmp	si,1FF8h
        je      Bank8SinCambio
        mov     [RomBank],1000h         ; bank 1 seleccionado
Bank8SinCambio:
	add	si,[RomBank]
WBRet:	ret
					        
; Banco de 16k

WBank16:and     si,01fffh
        cmp     si,1000h                ; si no es ROM, escribe de hardware
	jb	WriteHardware
        cmp     si,1FF6h                ; si no est en el rea de cambio (switch)...
        jb      Bank16SinCambio         ; no hay cambio
	cmp	si,1FFAh
        ja      Bank16SinCambio

        push    ax                      ; switch de ROMs de 16k
	mov	ax,si
        sub     ax,1FF6h                ; es banco 0,1,2,3
        shl     ax,12                   ; es 0,4k,8k,12k
	mov	[RomBank],ax
	pop	ax

Bank16SinCambio:
	add	si,[RomBank]
	ret



; *****************************************************************************
;     Mapeo de Memoria - Escritura - El compienzo SI contiene la direccin y
;     [WByte el dato. WSYNC y VSYNC son manejados aqu ya que afectan el timing
;     de la CPU.
; *****************************************************************************

WriteHardware:
        cmp     si,0400h                ; si es > 400h no escribe (es ROM)
	jge	WHret
        cmp     si,0200h                ; escribiendo al RIOT ?
	jge	WriteRIOT
WriteHardwarePage0:
	and	si,0FFh
        cmp     si,080h                 ; escribiendo al TIA ?
	jl	WriteTIA
        add     si,offset RiotRam-128   ; escribiendo a la RAM
	mov	dl,[WByte]
	mov	[si],dl
WHret:	ret

WriteRIOT:
        and     si,02FFh                ; puede ser en la direcciones 2xx  3xx
        cmp     si,0290h                ; escribiendo al timer (0290-029F)
        jl      WHret                   ; ir al cdigo que lo hace (ejecuta)!
	jmp	SetTimer
        

WriteTIA:
	and	si,03Fh
        cmp     si,04h                  ; escribiendo a registros de sincro.
        jl      WrTiaImmediate          ;   si, tratarlo inmediatamente
        cmp     si,02Ch                 ; verifica rango
        jg      XTRet                   ; todos los otros comando del TIA son
        push    bx                      ;   escritos a un buffer que contiene los
        push    ax                      ;   cambios al TIA en forma de registros (4 bytes)
	mov	al,[RClock]
	add	al,al
	add	al,[RClock]		; clock * 3
        sub     al,CycleOffset[si]      ; - 68 mas o menos (lugar de HBlank)
        ja      XTIANotHBlank           ; si es < 0 es un rea HBlank, ret 0
        xor     al,al                   ; color clock = 0
XTIANotHBlank:
        cmp     al,159                  ; debe ser < 160
        jb      XTIAOffRHS              ; est en rea de escritura ?
        mov     al,159                  ; si se pasa, ahora vale el mximo
XTIAOffRHS:
        mov     bx,[TIAWritePointer]    ; bx = primer lugar libre en buffer
        mov     TIAClock[bx],al         ; 1er pos: clock en que se grab
	mov	al,[WByte]	        
        mov     TIAVal[bx],al           ; 2da pos: valor del reg. siendo escrito
	mov	ax,si
        mov     TIAReg[bx],al           ; 3er pos: registro siendo modificado
	add	bx,4
        mov     [TIAWritePointer],bx    ; acutaliza primera posicin libre
        mov     byte ptr TIAClock[bx],TIAFin ; marca final en buffer
	pop	ax
	pop	bx
XTRet:	ret


WrTiaImmediate:
        shl     si,1                    ; perpara para la llamada
        jmp     [WrTiaImmediateVector + SI] ; llamada para ateneder inmedianente



; *****************************************************************************
;     A continuacin se tiene el cdigo de los registros que son tratados
;     inmediatamente, apendas son detectados en la ejecucin del cdigo emulado
; *****************************************************************************

;
; WSync -- espera por el borde princial del barrido horizontal
;

; este cdigo corrige el timer realizado llamadas a [TimerVec] para todos los
; ciclos restantes antes de VSYNC
        
WrTiaWSync:
        call    [TimerVec]              ; timing para la instruccin actual
        mov     bl,CiclosPorLineaBarrido; nmero de ciclos restantes en la lnea
        sub     bl,byte ptr [RClock]    ; nmero de clocks restantes para hacer
        js      WSret                   ; (fu js) si clock > 76 no hace nada
WSyncLoop:
        cmp     bl,8                    ; mx. de 8 ciclos de una.
        jl      WSyncExit               ; si es < 8 hace los ltimos que quedan
        mov     [RCycles],8             ; 'gasta' los 8 ciclos
	push	bx
	call	[TimerVec]
	pop	bx
        sub     bl,8                    ; resta 8 clocks ms
	jmp	WSyncLoop

WSyncExit:
        mov     [RCycles],bl            ; hace los ultimos ciclos en la llamada
                                        ; a [Timervec] luego del return
        mov     [RClock],CiclosPorLineaBarrido
WSret:  ret                             ; estamos al final de la lnea
                                        ; engaa al ciclo de ejecucin de scan
                                        ; terminandolo.



;
; VSync -- vertical sync set-clear
;

WrTiaVSync:     
        test    [WByte],2               ; si d1 est encendido entonces ....
	jz	WSret
NewFrame:
        mov     [LineaBarrido],0        ; vuelve al principio de la pantalla
;       mov     [RClock],0              ; por ahora lo dejamos fuera. Quilombo!
	call	TIABlank
        inc     [Frame]                 ; nuevo frame
	Ret
        


;
; VBlank -- vertical blank set-clear
;

WrTiaVBlank:
	push	ax
	mov	bl,[WByte]
	mov	[VBlank],bl
        test    bl,2                    ; encendiendo o apagando ?
        jz      WVBApagado              ;   apagando...
        mov     [LineaSup],65535        ; apaga el TIA
        jmp     WVBSale

WVBApagado:
        mov     ax,[LineaBarrido]
        cmp     ax,[CFirst]             ; LineaBarrido = primera lnea a escribir ?
        jae     WVBPasoMin
	mov	ax,[CFirst]
WVBPasoMin:
        mov     [LineaSup],ax
        add     ax,[MaxLineas]
        mov     [LineaInf],ax
	mov	al,[VBlankFrame]
	cmp	al,0
        jz      WVBSale
        inc     [Frame]
WVBSale:
	pop	ax
        ret
