; *****************************************************************************
;
;                       MACROS DE EMULACION PARA EL 6502
;
; *****************************************************************************
;
; Todas las lecturas y escrituras se hacen va READ/WRITE
;
;


read    macro   op1,op2                 ; lee la memoria en #2 y lo pone en #1
        
        mov     si,op2
        call    [ReadBank]              ; llama a la funcin de lectura del banco
        mov     op1,byte ptr [si]       ; graba resultado en #1

        endm
        

readsx  macro   op1,op2                 ; lee la memoria en #2 y lo pone en #1
                                        ; signo extendido
        mov     si,op2
        call    [ReadBank]              ; llama a la funcin de lectura del banco
        movsx   op1,byte ptr [si]       ; graba resultado en #1

        endm

read16  macro   op1,op2                 ; lee la memoria en #2 y lo pone en #1
        mov     si,op2                  ; lee 16 bits directamente en vez
        call    [ReadBank]              ;     de a un byte por vez
        mov     op1,word ptr [si]
        endm


write   macro   op1,op2                 ; escribe #1 en la pos. de memoria #2
        mov     si,op2                  ; SI = direccin a escribir
        mov     [WByte],op1             ; graba en WByte el valor a ser escrito
        call    [WriteBank]             ; llama a la funcin de escritura del banco
        endm
        
        
rewrite macro   op1                     ; reescribe el byte en #1
        mov     byte ptr [si],op1       ; asume SI como la direccin
        endm


zread   macro   op1,op2                 ; lectura de pgina cero
        mov     si,op2                  ; SI = direccin de lectura
        call    ReadHardwarePage0       ; lecutra de TIA o RAM
        mov     op1,[si]                ; op1 = byte ledo
        endm
        

zread16 macro   op1,op2                 ; lectura de pgina cero ver. 16 bits
        mov     si,op2                  ; SI = direccin de lectura
        call    ReadHardwarePage0       ; lectura de TIA o RAM
        mov     op1,[si]                ; op1 = byte ledo
        endm


zwrite  macro   op1,op2                 ; escritura de pgina cero
        mov     si,op2                  ; SI = direccin de escritura
        mov     [WByte],op1             ; graba en WByte el valor a ser escrito
        call    WriteHardwarePage0      ; escritura de TIA o RAM
        endm


fetch   macro   op1
        read    op1,bp                  ; lee byte apuntado por el PC
        inc     bp                      ; actualiza PC una vez ledo
        endm


fetchsx macro   op1
        readsx  op1,bp                  ; lee byte con signo apuntado por el PC
        inc     bp                      ; actualiza PC una vez ledo
        endm


fetch16 macro   op1
        read16  op1,bp                  ; lee word apuntado por el PC
        add     bp,2                    ; actualiza PC una vez ledo
        endm


IFDEF memcheats
absolute macro
        fetch16 bx                      ; BX = palabra apuntada por el PC
        endm
ELSE
absolute macro
        fetch   bl                      ; BL = primer byte apuntado por el PC
        fetch   bh                      ; BH = segundo byte apuntado por el PC
        endm
ENDIF   


zeropage macro                          ; hace un fetch de pgina cero --> BX
        fetch   bl                      ; BX = byte apuntado por el PC
        xor     bh,bh                   ; BH = 0
        endm


_index macro    op1
        add     bx,word ptr[op1]        ; BX = BX + [op1]
        endm


indexx  macro                           ; indice X
        _index  RX                      ; BX = BX + [X]
        endm
        

indexy  macro                           ; indice Y
        _index  RY                      ; BX = BX + [Y]
        endm


pagechange macro
        local   M1
        
        cmp     dh,bh                   ; cambi la pgina ?
        jz      M1                      ; no, sale
        inc     [RClock]                ; si, se necesita un ciclo extra
        inc     [RCycles]
M1:
        endm

IFDEF memcheats
readaddress macro                       ; lee dato en BX --> BX (debe estar en pgina cero)
        xor     bh,bh                   ; BH = 0. BX = direccin a leer
        read16  dx,bx                   ; DX = dato apuntado por BX
        mov     bx,dx                   ; BX = DX
        endm
ELSE
readaddress macro                       ; lee dato en BX --> BX (debe estar en pgina cero)
        xor     bh,bh                   ; BX = 0. BX = direccin a leer
        read    dl,bx                   ; lee el primer byte
        inc     bl
        read    dh,bx                   ; lee el segundo byte
        mov     bx,dx                   ; BX = DX
        endm
ENDIF   


absolutex macro                         ; absolute,x EAC
        absolute                        ; lee direccin apuntada por BP
        mov     dx,bx                   ; DX = dato ledo
        indexx                          ; BX = BX + [X]
        endm


absolutey macro                         ; absolute,y EAC
        absolute                        ; lee direccin apuntada por BP
        mov     dx,bx                   ; DX = dato ledo
        indexy                          ; BX = BX + [Y]
        endm


zeropagex macro                         ; fetch --> bl, y luego bl+[RX] (EAC)
        zeropage                        ; BL = byte ledo pgina cero, BX = 0
        add     bl,[RX]                 ; BL = BL + [X]
        endm


zeropagey macro                         ; fetch --> bl, y luego bl+[RY] (EAC)
        zeropage                        ; BL = byte ledo pgina cero, BX = 0
        add     bl,[RY]                 ; BL = BL + [Y]
        endm


indirectx macro                         ; (indirect,x) EAC
        zeropage                        ; BL = byte ledo pgina cero, BX = 0
        indexx                          ; BX = BX + [X]
        readaddress                     ; BX = [BX]
        endm


indirecty macro                         ; (indirect),y EAC
        zeropage                        ; BL = byte ledo pgina cero, BX = 0
        readaddress                     ; BX = [BX]
        mov     dx,bx                   ; DX = BX
        indexy                          ; BX = BX + [Y]
        endm


relative macro                          ; hace fetch de un byte con signo -> DX
        fetchsx dx                      ; DX = byte con signo ledo apuntado por BP
        endm


useztest macro  op1                     ; para test de Z
        mov     [RZTest],op1            ; actualiza Z
        endm
        

usentest macro  op1                     ; para test de N
        mov     [RNTest],op1            ; actualiza N
        endm
        

usetest macro   op1                     ; para test de N y Z (normal)
        useztest op1                    ; actualiza Z
        usentest op1                    ; actualiza N
        endm

; carga los registros de CPU

CargaRegs macro
        mov     ax,word ptr [RA]        ; mov al,[RA]; mov ah,[RCarry]
        mov     cl,[RS]                 ; CL = SP
        mov     bp,[RPC]                ; BP = PC
        endm

; graba los registro de CPU

SalvaRegs macro
        mov     word ptr [RA],ax        ; mov [RA],al; mov [RCarry],ah
        mov     [RS],cl                 ; SP = CL
        mov     [RPC],bp                ; PC = BP
        endm
