; *****************************************************************************
;
;                       EL CODIGO PRINCIPAL COMIENZA AQUI
;
; *****************************************************************************

comienzo:

        mov     ax,_TEXT                ;
        mov     ds,ax                   ; configura segmento de datos
        mov     bx,es                   ;
        mov     psp,bx                  ; graba puntero a PSP
        mov     es,ax                   ; configura el segmento extra
        cli                             ;
        mov     ss,ax                   ; configura segmento stack
        mov     sp,offset mystack       ;  y stack pointer
	sti
        
        mov     [CartRom+0FFCh],0       ; pone la direccin de reset por defecto
        mov     [CartRom+0FFDh],0f0h    ; a $F000

        call    LineaComandos           ; Procesa la lnea de comandos
        call    Reset                   ; Resetea la CPU
        SalvaRegs                       ; Graba el estado de la CPU
        call    TIAModoGrafico          ; Cambia a modo VGA
        call    SetNewInt9              ; Controlador de teclado propio


;*
;* Gran ciclo principal... "Pensar que un chip de ste tamao pueden
;* entrar millones de palabras".... (jejeje)
;*

ciclo_principal:
        test    [SinRetrace],1          ; Espera por retrace (slo VGA) ?
        jnz     VRterminado             ;  no
	
        mov     dx,03dah                ; espera por vertical retrace
esperaVR:
	in	al,dx
	test	al,8
        jnz     VRterminado
        jmp     esperaVR

VRterminado:
	call	ScanFrame	
pausa:  call    [CtrlFn]                ; verifica que teclas estn presionadas
        test    [JuegoPausado],1        ; espera mientras el juega est pausado
        jnz     pausa
        jmp     ciclo_principal

;*
;* fin del programa (se presion escape u opcode invlido)
;*

IrDOS:
        call    TIAModoTexto            ; vuelve a modo texto
        call    SetOldInt9              ; restaura el controlador de teclado

                                        ; vaca el buffer de teclado
Vacio:  mov     ah,0Bh                  ; teclado esperando....
	int	MSDOS
        cmp     al,0FFh                 ; sino, sale
        jne     Salir
        call    ConIn                   ; la lee
        jmp     Vacio
Salir:
        mov     ah,04Ch                 ; y retorna a MSDOS
	int	MSDOS

;*
;* hace algunas instrucciones hasta que RClock >= al
;*

HacerInstrucciones:
        mov     [UltimoCiclo],al        ; ltimo ciclo para hacer
        CargaRegs                       ; carga los registros de la CPU

OtraInstruccion:
        fetch   bl                      ; obtiene el operando (opcode)
        xor     bh,bh                   ;
        mov     dl,Cycles[bx]           ; obtiene #ciclos que le toma hacelo
        mov     [RCycles],dl            ; graba #ciclos
        add     [RClock],dl             ; agrega la #ciclos al reloj
        shl     bx,1                    ;
        jmp     [vectors + bx]          ; --> realiza la instruccin

ualnext:usetest al                      ; <-- instr. realizada, actualiza flags
next:   call    [TimerVec]              ; <-- instr. realizada, corre el reloj del RIOT
        mov     bh,[UltimoCiclo]        ; Hace un verificacin para saber
        cmp     [RClock],bh             ;    si se pas de mambo
        jl      OtraInstruccion         ;
        SalvaRegs                       ; graba los registros de CPU
	ret


;*
;* Hace un frame
;*

ScanFrame:
	mov	ax,0a000h
        mov     gs,ax                   ; apunta GS al rea de VGA
        mov     [DisplayPointer],0      ; resetea el puntero del display
ScanFrameLoop:
        mov     bx,[LineaBarrido]       ; resetea el TIA para una nueva lnea
        mov     [TIALineaBarrido],bx
	mov	bx,offset TIABuffer
	mov	[TIAWritePointer],bx
        mov     dword ptr [bx],-1       ;* puntero dword *

        call    TIALinea                ; genera una lnea (raster line)

        inc     [LineaBarrido]          ; Incrementa el contador de lineas
        sub     [RClock],CiclosPorLineaBarrido ; ajusta el reloj para la prx. linea

; testeo que la cantidad de lneas no se haya ido a la mierda y no est haciendo
; un VBlank.

        mov     al,[VBlankFrame]        ; VBlank ocasiona un nuevo frame
        cmp     al,0                    ; al = 0 ?
        jnz     NDret                   ;   si, saltea la prueba de emergencia
        mov     ax,[LineaBarrido]       ;
        cmp     ax,2000                 ; hicimos muchas lneas ?
        jl      NDret                   ;   todava no
        inc     [Frame]                 ;   si, al prxima frame (permite manejar el ESC)
        mov     [LineaBarrido],0        ; lnea actual = 0

NDret:  
	mov	ax,[Frame]
        cmp     ax,[FrameAnterior]      ; estamos en un nuevo frame ?
        je      ScanFrameLoop           ;   todava no
        mov     [FrameAnterior],ax      ;   si, lo marca como actual
        ret                             ;      y retorna

;*
;* pone en blanco el resto del display en cada frame
;*

TIABlank:
	push	ax
	push	di
	xor	ax,ax
	mov	di,[DisplayPointer]
TIABlankLoop:
	cmp	di,30000
        jb      TIABlankRet
        cmp     di,64000                ; alcanz el final del area de display ?
        jae     TIABlankRet             ;    si, listo!
	mov	gs:[di],ax
	add	di,2
        jmp     TIABlankLoop

TIABlankRet:
	mov	[DisplayPointer],di
	pop	di
	pop	ax
	ret

;*
;* aparece cuando hay un problem en el cdigo de operacion (opcode)
;*
  ErrMsg1 db 'Se ha producido un error en el cdigo.',13,13,10, 0
  ErrMsg2 db 'Cdigo de operacin: 0x',0
  ErrMsg3 db 'Direccin: 0x',0
  ErrMsg4 db 13,13,13,10,'Presione ESC para continuar...'
  ErrCrLf db 13,10,0
fail:
        mov     si, offset ErrMsg1
        call    ConStr
        mov     si, offset ErrMsg2
        call    ConStr
        dec     bp                      ; resguarda el puntero a la instruccin
        read    al,bp                   ; lee el byte
        call    ConHex                  ; lo muestra
        mov     si, offset ErrCrLf
        call    ConStr

        mov     si, offset ErrMsg3
        call    ConStr

        mov     ax,bp                   ; levanta la direccin
        call    ConWord                 ; la muestra

        mov     si, offset ErrMsg4
        call    ConStr

EsperaEsc:test    [TablaTecla+TeclaEsc],128   ; espera hasta que ESC sea presionada
        jz      EsperaEsc
        jmp     IrDOS


; *****************************************************************************
;
;                              Emulador del TIA
;
;       No preocuparse - los colores cambian an en el playfield
;
; *****************************************************************************

; *****************************************************************************
;
;       Continua la construccin de la linea hasta que la columna DH se alcanza
;       desde donde hemos dejado previamente (si los registros aun estn bien)
;
;       BP apunta al principio el buffer del TIA
;       DL prxima posicin de escritura del campo de juego
;       DH lmite de construccin de la lnea
;       SI prxima direccin de escritura del campo de juego
;
;       Este cdigo trabaja con el TIABuffer el cual es una lista de eventos,
;       sobre todo escrituras a los eventos del TIA, pero hay eventos especiales
;       para simular la ejecucin de RESxxx (escribir la lnea anterior) y
;       mltiples copias de grficos (como estar haciendo un RESxxx ficticio,
;       el cual no afecta el registro de posicin horizaontal).
;       Mezaclado con esto se encuentra la verificacin del bit del campo de juego,
;       el cual dibuja en bloques de 4 color clocks.
;       El cclo bsicamente avanza de a 4 color clocks. Trata de ejecutar todos
;       los eventos del TIABuffer si es que el clock del registro, es menor que
;       el clock en el que estamos. Luego, como el campo de juego tiene una
;       resolucin de 4 color clocks, se verifica en este momento para saber si
;       tiene un bit encendido. De ser as, 4 color clocks son dibujados en el
;       TIADisplay. El cclo continua hasta que se alcanza el final de la lnea,
;       en cuyo caso se procede a procesar una nueva lnea de barrido.
;
;       La lnea de pantalla es un array con registro de 6 bits (se utiliza un
;       byte por razones de comidad y simplicidad), cada uno representando
;       la presencia en la pantalla de objetos. Cuanto ms objetos en la misma
;       posicin halla, ms bit encendidos habr. (pf, bl, m0, p0, m1, p1)
;
;       En la entrada, se debe poner en el registro DH la posicin hasta donde
;       uno quiere dibujar. Luego, el sistema dibuja desde el comienzo hasta DH
;
; *****************************************************************************

;*
;* Macros para prueba de colisin
;*

CollisionTest macro op1
local Borra, Prox

        mov     bl,op1[si]              ; SI apunta a TIADisplay. Toma el byte #op1
        cmp     bl,3                    ; si es < 3 el valor, sale (no hay colisiones)
        jb      Prox
        shl     bl,1                    ; los datos en la tabla son WORDS
        jc      Borra                   ;   si el bit de reset de colisin est encendido...
        or      di,TIAColTab[bx]        ; di = di + nueva colisin
        jmp     Prox
        
Borra:
        xor     di,di                   ; borra primer la colisin
        or      di,TIAColTab[bx]        ; di = nueva colisin
Prox:

	endm

Collision4 macro
        mov     di,[TIACollide]         ; di = colisin acutal
        xor     bh,bh                   ; hace el test de colisin en los
                                        ;   prximos 4 bytes.
	CollisionTest 0
	CollisionTest 1
	CollisionTest 2
	CollisionTest 3

        mov     [TIACollide],di         ; copia el resultado al registro de colisin

	endm

;*
;* Macros de conversin de datos a salida en pantalla
;*

ConvertOut macro op1
        mov     bl,op1[si]              ; toma un byte del TIADisplay. SI = puntero a TIADisplay
        and     bx,03fh                 ; borra con una mscara los bits de colisin
        mov     di,[DisplayTranslate]   ;   slo quedan los bits que representan los objetos a dibujar
        mov     bl,[di+bx]              ; obtiene el color a dibujar de la tabla segn la tabla de prioridad.
                                        ;   el byte bx contiene todos los objetos al mismo tiempo en el byte a dibujar
        mov     bl,TIAColourTable[bx]   ; bl = obtiene el color de la tabla de colores
                                        ;    BK = posicin 0
                                        ;    PF = posicin 1
                                        ;    P1 = posicin 2
                                        ;    P0 = posicin 3
        mov     al,TIA[bx]              ; al = valor del color, segn el registro de color usado
        shr     al,1
        mov     ah,al                   ; dos bytes al mismo tiempo (160 clocks en pantalla deben ser 320)
        mov     di,[DisplayPointer]     ; di = puntero a posicin de escritura en pantalla
        mov     gs:[di],ax              ; escribe el byte
        add     di,2                    ; incrementa posicin
        mov     [DisplayPointer],di     ; actualiza puntero de escritura en pantalla
	endm


Convert4 macro
local	L1

	mov	ax,offset TIADisplayToColour
        test    [TIA+TIACtrlPf],4       ; si PF tiene prioridad,
        jz      L1                      ;    usa tabla de prioridad de color alternativa
	mov	ax,offset TIADisplayToColour2
L1:
        mov     [DisplayTranslate],ax   ; DisplyaTranslate = puntero a tabla de prioridades
        ConvertOut 0                    ; Dibuja el byte 0 desde la posicin en que se encuentra
        ConvertOut 1                    ;                1
        ConvertOut 2                    ;                2
        ConvertOut 3                    ;                3
	endm


;*
;* Macros para mostrar la salida del bit del campo de juego (playfield)
;*

BitPlayFieldIzquierdo macro
local  PFSkip, L2
                                        ; Obtiene el color correcto del campo de juego
        cmp     [NoScoreMode],0         ; Se est usando puntaje ?
	jnz	L2			;   no
	mov	[TIAColourTable+TCTPF],TIAColupf
        test    [TIA + TIACtrlPf],2     ; si hay separacin de color, color pf = col p0
        jz      L2
	mov	[TIAColourTable+TCTPF],TIAColup0

L2:                                     ; nota: un bit de pf = 4 color clocks
        mov     bx,cx                   ; bx = posicin verificada dentro del PF
        mov     eax,dword ptr [TIA+TIAPf0] ; eax = 4 bytes del PF
        test    eax,PfClockToBitTable[bx]  ; si el bit analizado est encendido ...
        jz      PFSkip                     ;   se escriben los 4 color clocks del PF en TIADisplay
	or	dword ptr [si],TGPlay*01000000h+TGPlay*010000h+TGPlay*0100h+TGPlay
PFSkip:
	endm


BitPlayFieldDerecho macro
local  PFSkip, L2                       
                                        ; Obtiene el color correcto del campo de juego
        cmp     [NoScoreMode],0         ; Se est usando puntaje ?
	jnz	L2			;   no
	mov	[TIAColourTable+TCTPF],TIAColupf
        test    [TIA + TIACtrlPf],2     ; si hay separacin de color, color pf = col p0
        jz      L2                      ;
	mov	[TIAColourTable+TCTPF],TIAColup1

L2:                                     ; nota: un bit de pf = 4 color clocks
        mov     bx,cx                   ; bx = posicin verificada dentro del PF
        mov     eax,dword ptr [TIA+TIAPf0] ; eax = 4 bytes del PF
        test    eax,PfClockToBitTable[bx]  ; si el bit analizado est encendido ...
        jz      PFSkip                     ;   se escriben los 4 color clocks del PF en TIADisplay
	or	dword ptr [si],TGPlay*01000000h+TGPlay*010000h+TGPlay*0100h+TGPlay
PFSkip:
	endm

;*
;* Macro para procesar objetos. verifica el clock en el que se est procesando,
;*    y si el clock del objeto alamacenado en el TIABuffer es menor que el clock
;*    en el que el emulador est situado, ejecuta el evento asociado al registro
;*    grabado en el TIABuffer
;*

HacerObjetos macro
local TIAObjetoLoop, TIASkip, TIAObjetoListo

TIAObjetoLoop:
        cmp     TIAClock[bp],dl         ; dl = clock acutal
        ja      TIAObjetoListo          ; si el clock del registro es > dl, luego escribe PF
        mov     bl,TIAReg[bp]           ;   bl = registro a ser procesado
        and     bx,03fh                 ; slo 63 opciones !!
	shl	bx,1
        call    TIARegVector[bx]        ; llama al evento asociado al registro
        
TIASkip:add     bp,4                    ; bp = prximo registro en TIABuffer
        cmp     byte ptr TIAClock[bp],TIAInvalido ; verifica si es un invlido
        je      TIASkip                 ; la saltea de ser as.
        jmp     TIAObjetoLoop
        
TIAObjetoListo:

	endm	

; *     
; *  TIALinea -- genera un barrido de lnea
; *
; *  Este es loop principal -- todo es tiempo crtico
; *

TIALinea:
        mov     bx,[LineaBarrido]
        cmp     bx,[LineaSup]           ; era 10
        jb      TIANoGenera
        cmp     bx,[LineaInf]           ; era 250
        jae     TIANoGenera

TIAGenera:
        mov     al,[CiclosPorLinea]     ; AL = cantidad de cclos por lnea
        call    HacerInstrucciones      ; ejecuta instrucciones mientras pueda

        mov     si,offset TIADisplay    ; puntero de escritura en campo de juego
        mov     bp,offset TIABuffer     ; inicializa el buffer

	mov	di,si
	xor	eax,eax
	mov	cx,40
        rep     stosd                   ; borra el buffer del TIADisplay

        call    TIAInsertaObjectos      ; inserta los objetos de la ltima lnea

        xor     dl,dl                   ; posicin de comienzo
        xor     cx,cx                   ; borra el clock del campo de juego

TIAIzquierdoLoop:                       ; se trabaja con bloques de 4 color clocks
        HacerObjetos                    ; ejecuta los registros del TIABuffer de ser necesario
        BitPlayFieldIzquierdo           ; escribe el campo de juego de ser necesario
	add	dl,4			; increment color clock
	add	cx,4			; increment playfield clock
	
        HacerObjetos                       ; ejecuta los registros del TIABuffer de ser necesario

        Convert4                        ; escribe 4 color clocks de del TIADisplay a pantalla

DoCollisionTest:
        Collision4                      ; actualiza registro de colisiones

        add     si,4                    ; incrementa el puntero de escritura del pf y de colisin
        cmp     dl,80                   ; alcanz el final de la lnea ?
        jae     TIADerechoEmpieza       ;   si, sale
        jmp     TIAIzquierdoLoop

TIADerechoEmpieza:
        test    [TIA+TIACtrlPf],1       ; invirtiendo la linea actual ?
        jnz     TIADerechoLoop          ;    si, continua
        xor     cx,cx                   ;    no, vuelve a comienza desde el principio

TIADerechoLoop:
        HacerObjetos                    ; ejecuta los registros del TIABuffer de ser necesario
        BitPlayFieldDerecho             ; escribe el campo de juego de ser necesario
        add     dl,4                    ; incrementa color clock
        add     cx,4                    ; incrementa clock del campo de juego
	
        HacerObjetos                    ; ejecuta los registros del TIABuffer de ser necesario
	Convert4
	Collision4

        add     si,4                    ; incrementa el puntero de escritura del pf y de colisin
        cmp     dl,160                  ; alcanz el final de la lnea ?
        jae     TIASale                 ;   si, sale
        jmp     TIADerechoLoop

TIASale:                                ; se termin la lnea de barrido...
	ret


;*
;* hace una lnea de barrido (fuera del area de display)
;*

TIANoGenera:
        mov     al,[CiclosPorLinea]
        call    HacerInstrucciones      ; ejecuta instrucciones mientras pueda

        mov     bp,offset TIABuffer     ; inicializa el TIABuffer
TIABLLoop:
        cmp     byte ptr TIAClock[bp],TIAFin    ; verifica si est al final del TIABuffer
        je      TIABLSale
        mov     bl,TIAReg[bp]           ; registro a cambiar
        and     bx,03fh                 ; slo 63 opciones !
	shl	bx,1
        call    TIARegVector2[bx]       ; ejecuta evento asociado a registro del TIAbuffer
        add     bp,4                    ; bp = apunta al prximo registro
	jmp	TIABLLoop
TIABLSale:
	ret


; *****************************************************************************
;
;     Rutinas de maneja del TIABuffer
;
;     Inserta todos los objetos (bl, p0, m0, p1, m1) de la ltima lnea en su
;     posicin actual como dice el registro del TIABuffer <lastclock> <reg+128> 0 0.
;     Esto hace que las mismas sean redibujadas en la prxima lnea en la misma
;     posicin. Si un objeto es redibujado antes de que sea alcanzado entonces
;     es envalidado.
;
; *****************************************************************************

TIAInsertaObjectos:
        mov     bx,TIAResp0             ; BX = offset en el TIA
TIAIOLoop:
        mov     al,TIA[bx]              ; AL = valor del clock
        mov     ah,bl                   ; BL = registro a ser escrito
        or      ah,128                  ; prende bit 7  (es una copias)
        call    TIAInserta              ; inserta el objeto
        inc     bx                      ; va al prximo objeto
        cmp     bx,TIAResbl             ; RESBL es el ltimo
	jle	TIAIOLoop
	ret

; *****************************************************************************
;
;     Ordena el buffer del TIA con respecto al campo 'Clock'
;
; *****************************************************************************

TIAOrdenaBuffer:
	push	ax
	push	bx
	push	cx
	push	dx
        mov     al,TIAClock[bp]         ; no ordena si hay 0  1 item
        cmp     al,TIAFin
        je      TIAOrdenaSalir
	mov	al,TIAClock+4[bp]
        cmp     al,TIAFin
        je      TIAOrdenaSalir
TIAOrdenaMain:
	mov	cl,0			; bl is set if any swaps this pass
        mov     bx,bp                   ; BX apunta dentro de la tabla
TIAOrdenaLoop:
        mov     al,TIAClock[bx]         ; obtiene el clock de las 2 prximas escrituras
	mov	ah,TIAClock+4[bx]
        cmp     al,ah                   ; si [primer] <= [segundo] evita el intercambio
        jbe     TIANoIntercambia

        mov     edx,[bx]                ; hace el intercambio
	mov	eax,4[bx]
	mov	[bx],eax
	mov	4[bx],edx

        inc     cl                      ; se ha realizado el intercambio
TIANoIntercambia:
        add     bx,4                    ; va al prximo par de escrituras
        mov     ah,TIAClock+4[bx]       ; si hay alguno para cambiar con el prx...
        cmp     ah,TIAFin
        jne     TIAOrdenaLoop           ; salta hacia atrs...

        cmp     cl,0                    ; se intercambi en este momento ?
        jnz     TIAOrdenaMain           ; si, vuelve y ordena nuevamente

TIAOrdenaSalir:
	pop	dx
	pop	cx
	pop	bx
	pop	ax
	ret


; *****************************************************************************
;
;     Inserta un registro en el buffer del TIA (apuntado por BP)
;                   AL = Valor del clock, AH = valor del registro
;
; *****************************************************************************

TIAInserta:
	push	bx
	push	cx
	push	bp
        push    ax                      ; este debe ser el ltimo....
        
TIAEncuentraPrimero:
        mov     bl,TIAClock[bp]         ; BL es el clock ac
	add	bp,4
        cmp     al,bl                   ; prueba el prx. clock si < clock insertado
        jae     TIAEncuentraPrimero

	sub	bp,4
        push    bp                      ; graba donde van los resultados

        mov     ebx,[bp]                ; primero para mandar adelante
TIAMandaAdelante:
	add	bp,4
        mov     ecx,[bp]                ; prxima a mandar adelante
        mov     [bp],ebx                ; graba que es lo que se envi hacia adelante
	mov	bl,TIAClock[bp]
        cmp     bl,TIAFin               ; fue el de marca final el que se envi ?
        je      TIAInserta2
        mov     ebx,ecx                 ; actualiza los registros de envo hacia adelante
        jmp     TIAMandaAdelante

TIAInserta2:
	pop	bp
        pop     ax                      ; restaura los datos y direcciones
	mov	TIAClock[bp],al
	mov	TIAReg[bp],ah
	xor	bl,bl
        mov     TIAVal[bp],bl           ; limpia el valor del campo

	pop	bp
	pop	cx
	pop	bx
        
        ret


;******************************************************************************
;
; Aqu yacen en paz los controladores de los registros del TIA
;
; Son llamados va la tabla vector durante la generacin de la lnea (TIALinea)
; Procesan los eventos del TIA (u objetos) que fueron encolados en el TIA buffer
; cuando se ejecut un grupo de instrucciones.
;
; La primera tabla vector (TIARegVector) es usada cuando la lnea est siendo dibujada
; mientras que la segunda (TIARegVector2) cuando la lnea est fuera del rea de dibujo.
;
; a la entrada tenemos:
;       bx = nmero del registro del TIA (lo cual implica bh = 0)
;       bp = puntero al evento del TIA en el TIA Buffer
;

.CEND
.DATASEGMENT

ALIGN 2

TIARegVector dw TRDummy         ; 00 -- VSYNC
        dw      TRDummy         ; 01 -- VBLANK
        dw      TRDummy         ; 02 -- WSYNC
        dw      TRDummy         ; 03 -- RSYNC
	dw	TRCopy		; 04 -- NUSIZ0
	dw	TRCopy		; 05 -- NUSIZ1
	dw	TRCopy		; 06 -- COLUP0
	dw	TRCopy		; 07 -- COLUP1
	dw	TRCopy		; 08 -- COLUPF
	dw	TRCopy		; 09 -- COLUBK
	dw	TRCopy		; 0A -- CTRLPF
	dw	TRCopy		; 0B -- REFP0
	dw	TRCopy		; 0C -- REFP1
	dw	TRCopy		; 0D -- PF0
	dw	TRCopy		; 0E -- PF1
	dw	TRCopy		; 0F -- PF2
	dw	TRPlyr		; 10 -- RESP0
	dw	TRPlyr		; 11 -- RESP1
	dw	TRMiss		; 12 -- RESM0
	dw	TRMiss		; 13 -- RESM1
	dw	TRBall		; 14 -- RESBL
        dw      TRDummy         ; 15 -- AUDC0 *
        dw      TRDummy         ; 16 -- AUDC1 *
        dw      TRDummy         ; 17 -- AUDF0 *
        dw      TRDummy         ; 18 -- AUDF1 *
        dw      TRDummy         ; 19 -- AUDV0 *
        dw      TRDummy         ; 1A -- AUDV1 *
	dw	TRGrp0		; 1B -- GRP0
	dw	TRGrp1		; 1C -- GRP1
	dw	TRCopy		; 1D -- ENAM0
	dw	TRCopy		; 1E -- ENAM1
	dw	TRCopy		; 1F -- ENABL
	dw	TRCopy		; 20 -- HMP0
	dw	TRCopy		; 21 -- HMP1
	dw	TRCopy		; 22 -- HMM0
	dw	TRCopy		; 23 -- HMM1
	dw	TRCopy		; 24 -- HMBL	
	dw	TRCopy		; 25 -- VDELP0
	dw	TRCopy		; 26 -- VDELP1
	dw	TRCopy		; 27 -- VDELBL
	dw	TRCopy		; 28 -- RESMP0
	dw	TRCopy		; 29 -- RESMP1
	dw	TRHMov		; 2A -- HMOVE
	dw	TRHClr		; 2B -- HMCLR
	dw	TRCClr		; 2C -- CXCLR

        dw      TRDummy         ; 2D
        dw      TRDummy         ; 2E
        dw      TRDummy         ; 2F
        dw      TRDummy         ; 30
        dw      TRDummy         ; 31
        dw      TRDummy         ; 32
        dw      TRDummy         ; 33
        dw      TRDummy         ; 34
        dw      TRDummy         ; 35
        dw      TRDummy         ; 36
        dw      TRDummy         ; 37
        dw      TRDummy         ; 38
        dw      TRDummy         ; 39
        dw      TRDummy         ; 3A
        dw      TRDummy         ; 3B
        dw      TRDummy         ; 3C
        dw      TRDummy         ; 3D
        dw      TRDummy         ; 3E
        dw      TRDummy         ; 3F



TIARegVector2 dw TRDummy        ; 00 -- VSYNC
        dw      TRDummy         ; 01 -- VBLANK
        dw      TRDummy         ; 02 -- WSYNC
        dw      TRDummy         ; 03 -- RSYNC
        dw      TRCopy          ; 04 -- NUSIZ0
	dw	TRCopy		; 05 -- NUSIZ1
	dw	TRCopy		; 06 -- COLUP0
	dw	TRCopy		; 07 -- COLUP1
	dw	TRCopy		; 08 -- COLUPF
	dw	TRCopy		; 09 -- COLUBK
	dw	TRCopy		; 0A -- CTRLPF
	dw	TRCopy		; 0B -- REFP0
	dw	TRCopy		; 0C -- REFP1
	dw	TRCopy		; 0D -- PF0
	dw	TRCopy		; 0E -- PF1
	dw	TRCopy		; 0F -- PF2
	dw	TRSave		; 10 -- RESP0
	dw	TRSave		; 11 -- RESP1
	dw	TRSave		; 12 -- RESM0
	dw	TRSave		; 13 -- RESM1
	dw	TRSave		; 14 -- RESBL
        dw      TRDummy         ; 15 -- AUDC0 *
        dw      TRDummy         ; 16 -- AUDC1 *
        dw      TRDummy         ; 17 -- AUDF0 *
        dw      TRDummy         ; 18 -- AUDF1 *
        dw      TRDummy         ; 19 -- AUDV0 *
        dw      TRDummy         ; 1A -- AUDV1 *
	dw	TRGrp0		; 1B -- GRP0
	dw	TRGrp1		; 1C -- GRP1
	dw	TRCopy		; 1D -- ENAM0
	dw	TRCopy		; 1E -- ENAM1
	dw	TRCopy		; 1F -- ENABL
	dw	TRCopy		; 20 -- HMP0
	dw	TRCopy		; 21 -- HMP1
	dw	TRCopy		; 22 -- HMM0
	dw	TRCopy		; 23 -- HMM1
	dw	TRCopy		; 24 -- HMBL
	dw	TRCopy		; 25 -- VDELP0
	dw	TRCopy		; 26 -- VDELP1
	dw	TRCopy		; 27 -- VDELBL
	dw	TRCopy		; 28 -- RESMP0
	dw	TRCopy		; 29 -- RESMP1
	dw	TRHmvB		; 2A -- HMOVE
	dw	TRHClr		; 2B -- HMCLR
	dw	TRCxZr		; 2C -- CXCLR

        dw      TRDummy         ; 2D
        dw      TRDummy         ; 2E
        dw      TRDummy         ; 2F
        dw      TRDummy         ; 30
        dw      TRDummy         ; 31
        dw      TRDummy         ; 32
        dw      TRDummy         ; 33
        dw      TRDummy         ; 34
        dw      TRDummy         ; 35
        dw      TRDummy         ; 36
        dw      TRDummy         ; 37
        dw      TRDummy         ; 38
        dw      TRDummy         ; 39
        dw      TRDummy         ; 3A
        dw      TRDummy         ; 3B
        dw      TRDummy         ; 3C
        dw      TRDummy         ; 3D
        dw      TRDummy         ; 3E
        dw      TRDummy         ; 3F
.DEND
.CODESEGMENT

;*
;* borra registro latch de colisin (versin on-screen)
;*

TRCClr: mov     bl,TIAClock[bp]             ; borra el flag de colisin
        or      byte ptr TIADisplay[bx],80h ; prende el bit 7 en el display field
        ret                                 ;

;*
;* borra registro latch de colisin (versin on-screen)
;*

TRCxZr: mov     [TIACollide],0          ; resetea la palabra de colisin
	ret

;*
;* no hace nada
;*
	        
TRDummy: ret                            ; simplemente retorna

;*
;* Copia registro. El registro del TIA obtiene el valor asignado en TIAVal
;*

TRCopy: mov     bl,TIAReg[bp]           ; bl = registro
                                        ; bh estaba en 0
        mov     al,TIAVal[bp]           ; al = valor
        mov     TIA[bx],al              ; lo escribe
	ret

;*
;* borra los registros de movimiento horizantal
;*    HMP0, HMP1, HMM0, HMM1, HMBL
;*

TRHClr: xor     al,al                   ; borra los registros HMove
	mov	[TIA+TIAHmp0],al
	mov	[TIA+TIAHmp1],al
	mov	[TIA+TIAHmm0],al
	mov	[TIA+TIAHmm1],al
	mov	[TIA+TIAHmbl],al
	ret

;*
;* grfico del player 0. Ver especificacin de vertical delay para entender
;* esta chanchada.
;*

TRGrp0: mov     al,TIAVal[bp]           ; actualiza registro de grfico 0
	mov	[TIA+TIAGrp0],al
        mov     al,[TIA+TIAGrp1]        ; copia Grp1 a Vertical Delay del Grp1
	mov	[VDP1Gr],al
	ret

;*
;* grfico del player 1
;*

TRGrp1: mov     al,TIAVal[bp]           ; actualiza registro de grfico 1
	mov	[TIA+TIAGrp1],al
        mov     al,[TIA+TIAGrp0]        ; copia Grp1 a Vertical Delay del Grp1
        mov     [VDP0Gr],al
	mov	al,[TIA+TIAEnabl]
	mov	[VDBlEn],al
	ret

; *
; * Ajusta posicin si se desfas
; *

TIAWrap macro
local TWneg,TWret
        cmp     al,240                  ; es negativo ???
        ja      TWneg                   ;   si, ajustar
        cmp     al,160                  ; pas la parte derecha ?
        jb      TWret                   ;   no
        sub     al,160                  ;   si
        jmp     TWret

TWneg:  xor     al,al                   ; ajustar
TWret:
	endm


;*
;* Aplica el movimiento horizantal (versin on-screen)
;*
;* Ajusta posiciones viejas (previamente escritas) de los registros HMxx
;*

TRHMov: xor     bx,bx                   ; bx indica el registro que se trabaja
                                        ; (empieza contando desde cero)
                                        ; (registros importantes empiezan en 10H)
                                        ; (por esto, se suma 90H en vez de 80H)
TRHMLoop:
        mov     al,bl                   ; obtiene el registro del TIABuffer a buscar
        add     al,090h                 ; obtiene su posicin real
        mov     ah,TIA+TIAHmp0[bx]      ; esta es la cantidad a mover
        sar     ah,4                    ; preserva el signo
        call    TIAActualizaPosicion    ; actualiza la posicin
        inc     bx                      ; hay 5 objetos para hacer.
	cmp	bx,5
	jne	TRHMLoop
        call    TIAOrdenaBuffer           ; el orden pudo ser afectado --> ordena
	ret

TIAActualizaPosicion:
	push	ax
	push	bp
        mov     bp,offset TIABuffer     ; BP = offset del TIABuffer
TIAAP2:
        cmp     byte ptr TIAClock[bp],TIAFin    ; verifica si est al final del buffer
        je      TIAAP1
	add	bp,4
        cmp     TIAReg-4[bp],al         ; verifica si encontr un registro de movimieto
        jne     TIAAP2                  ; previamente en el TIA buffer
        sub     bp,4                    ; bp apunta a l
        mov     al,TIAClock[bp]         ; AL = clock en que se gener
        sub     al,ah                   ; resta a la posicin el valor de movimiento
        TIAWrap                         ; ajusta posicin
        mov     TIAClock[bp],al         ; graba el resultado
TIAAP1: pop     bp
	pop	ax
        ret                             ; retorna


;*
;* Aplica el movimiento horizantal (versin off-screen)
;*

TRHmvB: mov	bx,0
TRHMB1: mov     ah,TIA+TIAHmp0[bx]      ; obtiene el ajuste
        sar     ah,4                    ; mantiene el signo
	mov	al,TIA+TIAResp0[bx]
        sub     al,ah                   ; lo ajusta !
	TIAWrap
	mov	TIA+TIAResp0[bx],al
	inc	bx
	cmp	bx,5
	jne	TRHMB1
	ret

;*
;* Copia el valor del clock en el registro. Registro = valor de clock
;*

TRSave: mov     bl,TIAReg[bp]           ; BX = offset del registro
        mov     al,TIAClock[bp]         ; AX = clock
	mov	TIA[bx],al
	ret
        


; *****************************************************************************
;
;                                Maneja la bola, chico!
;
; *****************************************************************************

TRBall:                                 ; hay que dibujar la bocha
        mov     bl,[TIA+TIAVDelbl]      ; mira el vertical delay de la bocha
        mov     al,[TIA+TIAEnabl]       ; AL = Flag de Enable
	test	bl,1
	jz	TRBBall2
        mov     al,[VDBlEn]             ; AL = Flag de Enable (vertical delay)
TRBBall2:
        test    al,2                    ; se fija si la bocha est prendida...
        jz      TRBNoDibuja             ; si es 0, se hace el otario

        mov     bl,TIAClock[bp]         ; toma el valor del clock en que se escribi el reg.
        xor     bh,bh                   ; limpia bh; en bx el valor del clock
        add     bx,offset TIADisplay    ; BX = comienzo del dibujo de la pelota en la lnea
        mov     al,[TIA+TIACtrlPf]      ; bits 5 y 4 controlan el tamao (CTRLPF)
        mov     ah,TGBall               ;
        call    TIAObjetoSolido         ; dibuja el objeto
                                        ; AH = bit que identifica al objeto siendo trabajado
                                        ; AL = Tamao en clocks
                                        ; BX = posicin en DisplayLine
TRBNoDibuja:
        mov     al,TIAClock[bp]         ; AL = clock actual
        mov     [TIA+TIAResbl],al       ; Actualiza RESbl. RESbl = clock
        mov     al,TIAReg[bp]           ; AL = Registro del record
        test    al,128                  ; si el bit 7 no est encendido (nuevo Reset...)
	jnz	TNret
        or      al,128                  ; no tena el bit 7 encendido
        call    TIAInvalida             ; invalda si hay copias viejas
TNret:	ret

; *****************************************************************************
;
;                                Control del Misil
;
;            BP = Puntero al registro donde est el misil en TIABuffer
;                               (clock/addr/data/0)
;
; *****************************************************************************

TRMiss:                                 ; hay que dibujar las balas
        mov     bl,TIAReg[bp]           ; si m0 ent bx = 0
        and     bx,1                    ; si m1 ent bx = 1

        mov     al,TIA+TIAResmp0[bx]    ; si la bala est en el centro del jugador,
        test    al,2                    ; no hace nada!
	jnz	TNret

        mov     al,TIA+TIAEnam0[bx]     ; se fija en el flag de enable para dibujar
        test    al,2                    ; si es 0, no dibuja
        jz      TRMNoDibuja

        mov     al,TIA+TIANuSiz0[bx]    ; AL = tamao del misil (bits 4,5)
        mov     ah,TGMissile0           ; AH = registro referencia del misil 0
        cmp     bl,0                    ; ve de que misil se trata
        jz      TGMCDibuja
        mov     ah,TGMissile1           ; AH = registro referencia del misil 1
TGMCDibuja:
	push	bx
        mov     bl,TIAClock[bp]         ; obtiene el clock (bh estaba en 0)
        add     bx,offset TIADisplay    ; bx apunta al comienzo del dibujo del misil
        call    TIAObjetoSolido         ; y lo dibuja...
        pop     bx                      ; AH = bit que identifica al objeto siendo trabajado
                                        ; AL = Tamao en clocks
                                        ; BX = posicin en DisplayLine
TRMNoDibuja:
        mov     al,TIAClock[bp]         ; RESm# = clock
        mov     TIA+TIAResm0[bx],al     ; Modifica la posicin del misil (off-screen)
        mov     al,TIAReg[bp]
        test    al,128                  ; si el bit 7 est vaco (nuevo Reset...)
	jnz	TGret
	or	al,128
        call    TIAInvalida             ; invalda las otras copias
TGret:	ret


; *****************************************************************************
;
;       Encuentra el primer item en TIABuffer (BP) tal que el registro = al
;       y lo marca como invlido. Mira desde el registro actual en adelante
;
; *****************************************************************************

TIAInvalida:
        push    bp                      ; guarda posicin del registro actual
TIAI1:  mov     ah,TIAClock[bp]         ; se fija si es el ltimo registro (marca de final)
        cmp     ah,TIAFin
	je	TIAI2
        cmp     al,TIAReg[bp]           ; si lo encontr, lo marca
        je      TIAIMarca
        add     bp,4                    ; sino, el que sigue
	jmp	TIAI1
TIAIMarca:                              ; lo marca invlido.
        mov     byte ptr TIAClock[bp],TIAInvalido
TIAI2:
	pop	bp
	ret

; *****************************************************************************
;
;    Dibuja un objeto (slido) de tamao 1,2,4  8 clocks en la direccin BX
;    (puntero en el display). AH contiene el bit de registro que se quiere
;    grabar en el TIADisplay. El tamao en clocks debe estar en AL
;    Resumen:
;       AH = Registro (objeto a ser trabajado)
;       AL = Tamao (cantidad de clocks)
;       BX = Direccin en TIADisplay donde ser escrito el registro
;
; *****************************************************************************

TIAObjetoSolido:
        and     al,030h                 ; enmascara slo los bits de tamao
                                        ; 30h = 0011 0000 (bit 4 y 5 son el tamao)
        jnz     TIAOSNot1
        or      [bx],ah                 ; prende el bit para size = 0. Escribe un solo byte
	ret
TIAOSNot1:
        push    cx                      ; se escribe al TIADisplay de a Words (2 bytes)
        mov     cl,ah                   ; cx = version word (bit de registro en cada byte)
        mov     ch,ah                   ; se escriben el bit en n clocks
        or      [bx],cx                 ; para 2 clocks
	cmp	al,010h
	je	TSRet
        or      2[bx],cx                ; para 4 clocks
	cmp	al,020h
	je	TSRet
        or      4[bx],cx                ; para 8 clocks
	or	6[bx],cx
TSRet:  
	pop	cx
	ret
        

; *****************************************************************************
;
;                                 Dibuja al Jugador
;
;                   BP apunta al player record (clock/addr/data/0)
;
; *****************************************************************************

TRPlyr:
        push    cx                      ; A dibujar al chabonito !!!
        mov     bl,TIAReg[bp]           ; si p0 ent bx = 0,
        and     bx,1                    ; si p1 ent bx = 1

        mov     ah,TGPlayer0            ; y el pixel a enceder...
        cmp     bl,0                    ; que es diferente para los jugadores
        jz      TGPCDibuja
	mov	ah,TGPlayer1
TGPCDibuja:
        mov     al,TIA+TIANuSiz0[bx]    ; AH = el tamao
        mov     cl,TIAReg[bp]           ; si el registro indca que es una copia
        test    cl,64                   ;   del objeto (bit 6 encendido), luego
        jz      TGPCNoCopia             ;   no debe darle bola al NuSize# ya que
        xor     al,al                   ;   sino, podra generar nuevas copias.
                                        ;   al = size = 0
TGPCNoCopia:
        mov     ch,TIA+TIAGrp0[bx]      ; CH = grfico del jugador (8 bits)
        mov     cl,TIA+TIAVDelp0[bx]    ; se fija el flag de vertical delay
	test	cl,1
	jz	TGPCNoVDelay
        mov     ch,VDP0Gr[bx]           ; usa el vertical delay
TGPCNoVDelay:
        mov     cl,TIA+TIARefP0[bx]     ; CL = el flag de reflection
	push	bx
        mov     bl,TIAClock[bp]         ; BL = clock  (bh estaba en 0)
        add     bx,offset TIADisplay    ; pone en bx la posicin del comienzo del dibujo
                                        ; del player dentro del display line

	push	si
        call    TIADibujaObjeto         ; lo dibuja...
	pop	si
	pop	bx

        mov     al,TIAReg[bp]           ; se fija si se trata de una copia....
        test    al,64                   ; verifica si es un objeto clon (insertado en TIABuffer)
        jnz     TRPNoSetRes             ; no pone RESP# a partir del clock

        mov     al,TIAClock[bp]         ; actualiza el RESp# con el clock
        mov     TIA+TIAResp0[bx],al     ; con lo cual ser dibujado en la prx. lnea

; *** HMMMM... Tema misiles... ***

        mov     ah,TIA+TIAResmp0[bx]    ; si la bala est en el centro
        test    ah,2                    ; del jugador (se come la bala???)
        jz      TRPNoSetRes             ;
        add     al,4                    ; hay reset... aumento clock
        mov     TIA+TIAResm0[bx],al     ; grabo nuevo clock en TIAResm0[bx]

TRPNoSetRes:
        mov     al,TIAReg[bp]           ; registro que fu escrito...
        test    al,128                  ; si el 7 no est encendido (nuevo Res...)
        jnz     TRPret                  ; sale si tiene el bit 7
	or	al,128
        call    TIAInvalida             ; invalda las otras copias
TRPret:
	pop	cx
	ret

; *****************************************************************************
;
;        Dibuja el jugador
;        [Los registros del TIAbuffer hasta 3Fh son los registros originales del TIA
;        Si tiene el bit 6 encendido, implca que es un objeto insertado duplicado
;
;        CH = regisro de tamao/copia
;        AL = en la pantalla en BX
;        CL = Flag de refelccin, AH es el bit para encender
;
; *****************************************************************************


TIADibujaObjeto:
        and     al,7                    ; me quedo con los bits que me interesan
        jnz     TGODiferente            ;   si existe alguno...
TGOStd: 
        cmp     ch,0                    ; normal
        jz      TGORet                  ;   ret
        test    cl,8                    ; reflejo?
        jnz     TGORef                  ;   s
        ConstruyeComun 0
        ConstruyeComun 1
        ConstruyeComun 2
        ConstruyeComun 3
        ConstruyeComun 4
        ConstruyeComun 5
        ConstruyeComun 6
        ConstruyeComun 7
TGORet: ret
        
TGORef: ConstruyeInverso 0
        ConstruyeInverso 1
        ConstruyeInverso 2
        ConstruyeInverso 3
        ConstruyeInverso 4
        ConstruyeInverso 5
        ConstruyeInverso 6
        ConstruyeInverso 7
	ret
        

TGODiferente:                           ; Maneja diferencias.......
	movzx	esi,al
	jmp	[offset TGODiffTab + esi*2]

TGO3Close:                              ; 3 Copias, cerca (close) en clocks 0,16,32
        call    TGOStd                  ; Dibuja objeto (TIADisplay) en clock = 0
        mov     al,TIAClock[bp]         ; AL = clock
        mov     ah,TIAReg[bp]           ; AH = registro
        mov     cx,ax
        add     al,16                   ; AL = clock + 16
        or      ah,64                   ; seteo que es un un objeto clon (p/insertar en TIABuffer)
        call    TIAInserta              ; y lo inserta...
        xchg    ax,cx                   ; cx contena TIAClock y TIAReg anteriores
        add     al,32                   ; AL = clock + 32
        or      ah,64                   ; seteo que es un un objeto clon (p/insertar en TIABuffer)
        call    TIAInserta              ; y lo inserta...
	ret

TGO2Close:                              ; 2 Copias,cerca (close) separados por 16 ciclos
	call	TGOStd
	mov	al,TIAClock[bp]
	mov	ah,TIAReg[bp]
        add     al,16                   ; AL = clock + 16
        or      ah,64                   ; seteo que es un un objeto clon (p/insertar en TIABuffer)
        call    TIAInserta
	ret

TGO3Medium:                             ; 3 copias,ms_o_menos (medium) @ 0,32,64
	call	TGOStd
	mov	al,TIAClock[bp]
	mov	ah,TIAReg[bp]
	mov	cx,ax
        add     al,32                   ; AL = clock + 32
        or      ah,64                   ; seteo que es un un objeto clon (p/insertar en TIABuffer)
        call    TIAInserta
	xchg	ax,cx
        add     al,64                   ; AL = clock + 64
        or      ah,64                   ; seteo que es un un objeto clon (p/insertar en TIABuffer)
        call    TIAInserta
	ret

TGO2Medium:                             ; 2 copias, ms_o_menos (medium)
                                        ; seperados por 32 ciclos
	call	TGOStd
	mov	al,TIAClock[bp]
	mov	ah,TIAReg[bp]
        add     al,32                   ; AL = clock + 32
        or      ah,64                   ; seteo que es un un objeto clon (p/insertar en TIABuffer)
        call    TIAInserta
	ret

TGO2Far:                                ; 2 copias lejos (far) seperados por 64 ciclos
	call	TGOStd
	mov	al,TIAClock[bp]
	mov	ah,TIAReg[bp]
        add     al,64                   ; AL = clock + 64
        or      ah,64                   ; seteo que es un un objeto clon (p/insertar en TIABuffer)
        call    TIAInserta
	ret


TGO2Width:
        cmp     ch,0                    ; normal
        jz      TGORet                  ;   retorna
	mov	al,ah
        test    cl,8                    ; reflejo?
        jnz     TGO2Ref                 ;   s
        ConstruyeComun2 0
        ConstruyeComun2 1
        ConstruyeComun2 2
        ConstruyeComun2 3
        ConstruyeComun2 4
        ConstruyeComun2 5
        ConstruyeComun2 6
        ConstruyeComun2 7
	ret
        
TGO2Ref:
        ConstruyeInverso2 0
        ConstruyeInverso2 1
        ConstruyeInverso2 2
        ConstruyeInverso2 3
        ConstruyeInverso2 4
        ConstruyeInverso2 5
        ConstruyeInverso2 6
        ConstruyeInverso2 7
	ret

        
TGO4Width:
        cmp     ch,0                    ; normal
        jz      TGORet                  ;   ret
	mov	al,ah
	push	ax
	shl	eax,16
	pop	ax
        test    cl,8                    ; reflejado?
        jnz     TGO4Ref                 ;   s
        ConstruyeComun4 0
        ConstruyeComun4 1
        ConstruyeComun4 2
        ConstruyeComun4 3
        ConstruyeComun4 4
        ConstruyeComun4 5
        ConstruyeComun4 6
        ConstruyeComun4 7
	ret
        
TGO4Ref:
        ConstruyeInverso4 0
        ConstruyeInverso4 1
        ConstruyeInverso4 2
        ConstruyeInverso4 3
        ConstruyeInverso4 4
        ConstruyeInverso4 5
        ConstruyeInverso4 6
        ConstruyeInverso4 7
        ret

