; *****************************************************************************
;
;                     EMULACION DE TECLAS DE CONTROL
;
; Esta funcin es llamada una vez por frame para verificar es estado del teclado
;
; *****************************************************************************

; definiciones de teclas de control

TeclaEsc  =               001h            ; ESC (Sale al DOS)

TeclaF1   =               03Bh            ; F1  (Reset)
TeclaF2   =               03Ch            ; F2  (Seleccin)
TeclaF3   =               03Dh            ; F3  (no asignado)
TeclaF4   =               03Eh            ; F4  (no asignado)
TeclaF5   =               03Fh            ; F5  (no asignado)
TeclaF6   =               040h            ; F6  (no asignado)
TeclaF7   =               041h            ; F7  (no asignado)
TeclaF8   =               042h            ; F8  (no asignado)
TeclaF9   =               043h            ; F9  (Blanco y Nergo)
TeclaF10  =               044h            ; F10 (Color)

TeclaIzq =                04Bh            ; flecha izquierda
TeclaDer =                04Dh            ; flecha derecha
TeclaArr =                048h            ; flecha arriba
TeclaAbajo =              050h            ; flecha abajo
TeclaSpace =              039H            ; barra espaciador (disparo)

TeclaZ =                  02Ch            ; Z (izquierda 2do Jugador)
TeclaX =                  02Dh            ; X (derecha 2do Jugador)
TeclaC =                  02Eh            ; C (abajo 2do Jugador)
TeclaF =                  021h            ; F (arriba 2do Jugador)
TeclaG =                  022h            ; G (disparo 2do Jugador)

TeclaP =                  019H            ; P (pausa el juego)
TeclaEnter =              01CH            ; Enter (continua el juego)

Controls:
        mov     ah,[IOPortB]              ; enciende los bits de Seleccin y Reset
	or	ah,03h
        test    [TablaTecla+TeclaF1],128
        jz      Ctrl1                     ; F1 presionado (Reset) - borra bit 0 (1)
	and	ah,0FEh
Ctrl1:  test    [TablaTecla+TeclaF2],128
        jz      Ctrl2                     ; F2 presionado (Seleccin) - borra bit 1 (2)
	and	ah,0FDh
Ctrl2:  
        test    [TablaTecla+TeclaF9],128
        jz      Ctrl2a                    ; F9 presionado (Blanco y negro) borra bit 3 (8)
	and	ah,0F7h
Ctrl2a: test    [TablaTecla+TeclaF10],128
        jz      Ctrl2b                    ; F10 presionado (Color) enciende bit 3 (8)
	or	ah,08h
Ctrl2b: 
	mov	[IOPortB],ah

        mov     ah,0FFh                   ; ahora los switches

        TeclaUsada TeclaDer,128
        TeclaUsada TeclaIzq,64
        TeclaUsada TeclaAbajo,32
        TeclaUsada TeclaArr,16

        TeclaUsada TeclaX,8
        TeclaUsada TeclaZ,4
        TeclaUsada TeclaC,2
        TeclaUsada TeclaF,1

        mov     [IOPortA],ah              ; graba los switches del Puerto A

Ctrl5:	mov	[InputLatch],080h
	mov	[InputLatch+1],080h
CTLatching:

        test    [TablaTecla+TeclaSpace],128
        jz      Ctrl3                     ; maneja el 'disparo' del jugador 1
	mov	[InputLatch],0h
Ctrl3:  test    [TablaTecla+TeclaG],128
        jz      Ctrl4                     ; maneja el 'disparo' del jugador 2
	mov	[InputLatch+1],0h
Ctrl4:  test    [TablaTecla+TeclaP],128
	jz	Ctrl6
        mov     [JuegoPausado],1          ; pausa el juego
Ctrl6:  test    [TablaTecla+TeclaEnter],128
	jz	Ctrl7
        mov     [JuegoPausado],0          ; continua el juego
Ctrl7:  test    [TablaTecla+TeclaEsc],128
        jne     IrDOS                     ; ESC presionado - sale
CtrlExit:
	ret
	

; *****************************************************************************
;
;   Controlador de INT9 alternativo, permite controlar el estado del teclado
;
; *****************************************************************************


datasegsave     dw      0               ; copia del segmento de datos

SetNewInt9:
	push	es
	push	si
	push	bx
	push	ax
	mov	ax,ds
        mov     cs:datasegsave,ax       ; graba una copia del segmento de datos
	cli
        mov     ax,3509h                ; obtiene la vieja INT 9 (por ahora actual)
	int	21h
	mov	si,OldInt9
        mov     [OldInt9],bx            ; graba desplazamiento
        mov     [OldInt9+2],es          ; graba segmento
	push	ds
	push	cs
	pop	ds
        mov     ax,2509h                ; configura nueva INT 9
        mov     dx,offset NewInt9       ; DS:DX nuevo vector de interrupcin
	int	21h
	pop	ds
	sti
	pop	ax
	pop	bx
	pop	si
	pop	es
	ret

SetOldInt9:
	push	ds
	cli
        mov     ax,2509h                ; configura nueva INT 9
        mov     dx,[OldInt9]            ; carga desplazamiento
        mov     ds,[OldInt9+2]          ; carga segmento
	int	21h
	sti
	pop	ds
	ret


NewInt9:push	ax
	push	bx
	push	ds
	mov	ax,cs:datasegsave
	mov	ds,ax
        in      al,60h                  ; obtiene el scan code

        mov     bl,al                   ; BX = offset en la tabla
	and	bl,07Fh
	xor	bh,bh
        and     al,080h                 ; asla el bit on-off
        xor     al,080h                 ; bit 7 0. Si tecla presionada, la invierte
        mov     TablaTecla[bx],al       ; lo escribe en la tabla

        mov     al,20h
	out	20h,al
                                        ; (bit 7 encendido si es presionado)
	pop	ds				        
	pop	bx
	pop	ax
	sti
	iret
